import pytest
from faust import web
from faust.exceptions import ImproperlyConfigured
from faust.sensors.statsd import StatsdMonitor
from faust.types import TP
from mode.utils.mocks import Mock, call

TP1 = TP('foo', 3)


class test_StatsdMonitor:

    @pytest.fixture
    def time(self):
        timefun = Mock(name='time()')
        timefun.return_value = 101.1
        return timefun

    @pytest.fixture()
    def statsd(self, *, monkeypatch):
        statsd = Mock(name='statsd')
        monkeypatch.setattr('faust.sensors.statsd.statsd', statsd)
        return statsd

    @pytest.fixture()
    def stream(self):
        stream = Mock(name='stream')
        stream.shortlabel = 'Stream: Topic: foo'
        return stream

    @pytest.fixture()
    def event(self):
        return Mock(name='event')

    @pytest.fixture()
    def table(self):
        table = Mock(name='table')
        table.name = 'table1'
        return table

    @pytest.fixture()
    def req(self):
        return Mock(name='request', autospec=web.Request)

    @pytest.fixture()
    def response(self):
        return Mock(name='response', autospec=web.Response)

    @pytest.fixture()
    def view(self):
        return Mock(name='view', autospec=web.View)

    @pytest.fixture()
    def mon(self, *, statsd, time):
        return StatsdMonitor(time=time)

    def test_statsd(self, *, mon):
        assert mon.client

    def test_raises_if_statsd_not_installed(self, *, monkeypatch):
        monkeypatch.setattr('faust.sensors.statsd.statsd', None)
        with pytest.raises(ImproperlyConfigured):
            StatsdMonitor()

    def test_on_message_in_out(self, *, mon):
        message = Mock(name='message')
        mon.on_message_in(TP1, 400, message)

        mon.client.incr.assert_has_calls([
            call('messages_received', rate=mon.rate),
            call('messages_active', rate=mon.rate),
            call('topic.foo.messages_received', rate=mon.rate),
        ])
        mon.client.gauge.assert_called_once_with('read_offset.foo.3', 400)

        mon.on_message_out(TP1, 400, message)
        mon.client.decr.assert_called_once_with(
            'messages_active', rate=mon.rate)

    def test_on_stream_event_in_out(self, *, mon, stream, event):
        state = mon.on_stream_event_in(TP1, 401, stream, event)
        mon.client.incr.assert_has_calls([
            call('events', rate=mon.rate),
            call('stream.topic_foo.events', rate=mon.rate),
            call('events_active', rate=mon.rate),
        ])
        mon.on_stream_event_out(TP1, 401, stream, event, state)
        mon.client.decr.assert_called_once_with('events_active', rate=mon.rate)
        mon.client.timing.assert_called_once_with(
            'events_runtime',
            mon.secs_to_ms(mon.events_runtime[-1]),
            rate=mon.rate,
        )

    def test_on_table_get(self, *, mon, table):
        mon.on_table_get(table, 'key')
        mon.client.incr.assert_called_once_with(
            'table.table1.keys_retrieved', rate=mon.rate,
        )

    def test_on_table_set(self, *, mon, table):
        mon.on_table_set(table, 'key', 'value')
        mon.client.incr.assert_called_once_with(
            'table.table1.keys_updated', rate=mon.rate,
        )

    def test_on_table_del(self, *, mon, table):
        mon.on_table_del(table, 'key')
        mon.client.incr.assert_called_once_with(
            'table.table1.keys_deleted', rate=mon.rate,
        )

    def test_on_commit_completed(self, *, mon):
        consumer = Mock(name='consumer')
        state = mon.on_commit_initiated(consumer)
        mon.on_commit_completed(consumer, state)
        mon.client.timing.assert_called_once_with(
            'commit_latency', mon.ms_since(float(state)), rate=mon.rate,
        )

    def test_on_send_initiated_completed(self, *, mon):
        producer = Mock(name='producer')
        state = mon.on_send_initiated(
            producer, 'topic1', 'message', 321, 123)
        mon.on_send_completed(producer, state, Mock(name='metadata'))

        mon.client.incr.assert_has_calls([
            call('topic.topic1.messages_sent', rate=mon.rate),
            call('messages_sent', rate=mon.rate),
        ])
        mon.client.timing.assert_called_once_with(
            'send_latency',
            mon.ms_since(float(state)), rate=mon.rate,
        )

        mon.on_send_error(producer, KeyError('foo'), state)
        mon.client.incr.assert_has_calls([
            call('messages_sent_error', rate=mon.rate),
        ])
        mon.client.timing.assert_has_calls([
            call('send_latency_for_error',
                 mon.ms_since(float(state)), rate=mon.rate),
        ])

    def test_on_assignment_start_completed(self, *, mon):
        assignor = Mock(name='assignor')
        state = mon.on_assignment_start(assignor)
        mon.on_assignment_completed(assignor, state)

        mon.client.incr.assert_has_calls([
            call('assignments_complete', rate=mon.rate),
        ])
        mon.client.timing.assert_has_calls([
            call('assignment_latency',
                 mon.ms_since(state['time_start']), rate=mon.rate),
        ])

    def test_on_assignment_start_failed(self, *, mon):
        assignor = Mock(name='assignor')
        state = mon.on_assignment_start(assignor)
        mon.on_assignment_error(assignor, state, KeyError())

        mon.client.incr.assert_has_calls([
            call('assignments_error', rate=mon.rate),
        ])
        mon.client.timing.assert_has_calls([
            call('assignment_latency',
                 mon.ms_since(state['time_start']), rate=mon.rate),
        ])

    def test_on_rebalance(self, *, mon):
        app = Mock(name='app')
        state = mon.on_rebalance_start(app)

        mon.client.incr.assert_has_calls([
            call('rebalances', rate=mon.rate),
        ])

        mon.on_rebalance_return(app, state)

        mon.client.incr.assert_has_calls([
            call('rebalances', rate=mon.rate),
            call('rebalances_recovering', rate=mon.rate),
        ])
        mon.client.decr.assert_has_calls([
            call('rebalances', rate=mon.rate),
        ])
        mon.client.timing.assert_has_calls([
            call('rebalance_return_latency',
                 mon.ms_since(state['time_return']), rate=mon.rate),
        ])

        mon.on_rebalance_end(app, state)

        mon.client.decr.assert_has_calls([
            call('rebalances', rate=mon.rate),
            call('rebalances_recovering', rate=mon.rate),
        ])

        mon.client.timing.assert_has_calls([
            call('rebalance_return_latency',
                 mon.ms_since(state['time_return']), rate=mon.rate),
            call('rebalance_end_latency',
                 mon.ms_since(state['time_end']), rate=mon.rate),
        ])

    def test_on_web_request(self, *, mon, request, response, view):
        response.status = 404
        self.assert_on_web_request(mon, request, response, view,
                                   expected_status=404)

    def test_on_web_request__None_response(self, *, mon, request, view):
        self.assert_on_web_request(mon, request, None, view,
                                   expected_status=500)

    def assert_on_web_request(self, mon, request, response, view,
                              expected_status):
        app = Mock(name='app')
        state = mon.on_web_request_start(app, request, view=view)
        mon.on_web_request_end(app, request, response, state, view=view)

        mon.client.incr.assert_has_calls([
            call(f'http_status_code.{expected_status}', rate=mon.rate),
        ])

        mon.client.timing.assert_has_calls([
            call('http_response_latency',
                 mon.ms_since(state['time_end']), rate=mon.rate),
        ])

    def test_count(self, *, mon):
        mon.count('metric_name', count=3)
        mon.client.incr.assert_called_once_with(
            'metric_name', count=3, rate=mon.rate)

    def test_on_tp_commit(self, *, mon):
        offsets = {
            TP('foo', 0): 1001,
            TP('foo', 1): 2002,
            TP('bar', 3): 3003,
        }
        mon.on_tp_commit(offsets)
        mon.client.gauge.assert_has_calls([
            call('committed_offset.foo.0', 1001),
            call('committed_offset.foo.1', 2002),
            call('committed_offset.bar.3', 3003),
        ])

    def test_track_tp_end_offsets(self, *, mon):
        mon.track_tp_end_offset(TP('foo', 0), 4004)
        mon.client.gauge.assert_called_once_with(
            'end_offset.foo.0', 4004,
        )



#!/usr/bin/env python
import faust

app = faust.App(
    'hello-world',
    broker='kafka://localhost:9092',
)

greetings_topic = app.topic('greetings', value_type=str)


@app.agent(greetings_topic)
async def print_greetings(greetings):
    async for greeting in greetings:
        print(greeting)


@app.timer(5)
async def produce():
    for i in range(100):
        await print_greetings.send(value=f'hello {i}')

if __name__ == '__main__':
    app.main()



from itertools import count

import aredis
import pytest
import faust
from faust.exceptions import ImproperlyConfigured
from faust.web import Blueprint, View
from faust.web.cache import backends
from faust.web.cache.backends import redis
from mode.utils.mocks import Mock

DEFAULT_TIMEOUT = 361.363
VIEW_B_TIMEOUT = 64.3

blueprint = Blueprint('test')
cache = blueprint.cache(timeout=None)


class ResponseModel(faust.Record):
    key: str
    value: int


@blueprint.route('/A/', name='a')
class ACachedView(View):

    def __post_init__(self) -> None:
        self.counter = count()

    @cache.view(timeout=DEFAULT_TIMEOUT)
    async def get(self, request):
        return await self._next_response(request)

    async def _next_response(self, request):
        cache_key = cache.key_for_request(request, None, request.method)
        return self.json(ResponseModel(
            key=cache_key,
            value=next(self.counter),
        ))


@blueprint.route('/B/', name='b')
class BCachedView(ACachedView):

    @cache.view(timeout=VIEW_B_TIMEOUT)
    async def get(self, request):
        return await self._next_response(request)


@blueprint.route('/C', name='c')
class CCachedView(ACachedView):
    ...


@blueprint.route('/D/', name='d')
class DCachedView(ACachedView):

    @cache.view(timeout=None)
    async def get(self, request):
        return await self._next_response(request)


def test_cache():
    assert cache.key_prefix == 'test'


@pytest.mark.asyncio
@pytest.mark.app(cache='memory://')
async def test_cached_view__HEAD(*, app, bp, web_client, web):
    app.cache.storage.clear()
    async with app.cache:
        client = await web_client
        urlA = web.url_for('test:a')
        response = await client.head(urlA)
        assert response.status == 200
        response2 = await client.head(urlA)
        assert response2.status == 200


@pytest.mark.asyncio
@pytest.mark.app(cache='redis://')
async def test_cached_view__redis_no_timeout(
        *, app, bp, web_client, web, mocked_redis):
    async with app.cache:
        client = await web_client
        urlD = web.url_for('test:d')
        response = await client.get(urlD)
        assert response.status == 200
        content = await response.read()
        response2 = await client.get(urlD)
        assert response2.status == 200
        assert await response2.read() == content


@pytest.mark.asyncio
@pytest.mark.app(cache='memory://')
async def test_cached_view__memory_no_timeout(
        *, app, bp, web_client, web, mocked_redis):
    async with app.cache:
        client = await web_client
        urlD = web.url_for('test:d')
        response = await client.get(urlD)
        assert response.status == 200
        content = await response.read()
        response2 = await client.get(urlD)
        assert response2.status == 200
        assert await response2.read() == content


@pytest.mark.asyncio
@pytest.mark.app(cache='memory://')
async def test_cached_view__cannot_cache(*, app, bp, web_client, web):
    prev = cache.can_cache_request
    cache.can_cache_request = Mock(name='can_cache_request')
    cache.can_cache_request.return_value = False
    try:
        app.cache.storage.clear()
        async with app.cache:
            client = await web_client
            urlA = web.url_for('test:a')
            response = await client.get(urlA)
            assert response.status == 200
            response2 = await client.get(urlA)
            assert response2.status == 200
    finally:
        cache.can_cache_request = prev


@pytest.mark.app(cache='memory://')
@pytest.mark.parametrize('include_headers', [True, False])
def test_key_for_request(include_headers, *, app):
    _cache = blueprint.cache(timeout=DEFAULT_TIMEOUT)
    request = Mock(name='request')
    _cache.build_key = Mock(name='build_key')
    _cache.key_for_request(
        request, prefix='/foo/', include_headers=include_headers)
    _cache.build_key.assert_called_once_with(
        request,
        request.method,
        '/foo/',
        request.headers if include_headers else {},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize('expected_backend', [
    pytest.param('redis', marks=pytest.mark.app(cache='redis://')),
    pytest.param('memory', marks=pytest.mark.app(cache='memory://')),
])
async def test_cached_view__redis(expected_backend, *,
                                  app, bp, web_client, web, mocked_redis):
    cache_backend = app.cache
    if cache_backend.url.scheme == 'redis':
        storage = mocked_redis.storage
        timeout_t = int
    else:
        storage = app.cache.storage
        timeout_t = lambda timeout: timeout  # noqa
    async with cache_backend:
        assert cache_backend.url.scheme == expected_backend

        client = await web_client
        urlA = web.url_for('test:a')
        urlB = web.url_for('test:b')
        urlC = web.url_for('test:c')
        assert urlA == '/test/A/'
        assert storage.ttl('does-not-exist') is None
        responseA = await model_response(await client.get(urlA))
        keyA = responseA.key
        assert '.test.' in keyA
        assert responseA.value == 0
        assert await model_value(await client.get(urlA)) == 0
        assert storage.get(keyA)
        assert storage.ttl(keyA)
        assert (timeout_t(storage.last_set_ttl(keyA)) ==
                timeout_t(DEFAULT_TIMEOUT))

        assert await model_value(await client.get(urlA)) == 0

        storage.expire(keyA)
        assert await model_value(await client.get(urlA)) == 1
        assert await model_value(await client.get(urlA)) == 1
        assert await model_value(await client.get(urlA)) == 1

        responseB = await model_response(await client.get(urlB))
        keyB = responseB.key
        assert '.test.' in keyB
        assert responseB.value == 0
        assert await model_value(await client.get(urlB)) == 0
        assert await model_value(await client.get(urlB)) == 0
        assert await model_value(await client.get(urlB)) == 0
        assert (timeout_t(storage.last_set_ttl(keyB)) ==
                timeout_t(VIEW_B_TIMEOUT))
        storage.expire(keyB)
        assert await model_value(await client.get(urlB)) == 1
        storage.expire(keyB)
        assert await model_value(await client.get(urlB)) == 2
        storage.expire(keyB)
        assert await model_value(await client.get(urlB)) == 3

        # A is now 1
        # B is at 3
        # C should still be at 0
        assert await model_value(await client.get(urlC)) == 0
        assert await model_value(await client.get(urlC)) == 0

        storage._expires['xuzzy'] = None
        storage._time_index['xuzzy'] = None
        assert storage.get('xuzzy') is None
        assert 'xuzzy' not in storage._expires
        assert 'xuzzy' not in storage._time_index


@pytest.mark.asyncio
@pytest.mark.parametrize('scheme,host,port,password,db,settings', [
    pytest.param('redis', 'h', 6379, None, 0, None,
                 marks=pytest.mark.app(cache='redis://h:6379')),
    pytest.param('redis', 'h', 6379, None, 0, None,
                 marks=pytest.mark.app(cache='redis://h:6379/')),
    pytest.param('redis', 'h', 6379, None, 0, None,
                 marks=pytest.mark.app(cache='redis://h:6379/0')),
    pytest.param('redis', 'h', 6379, None, 1, None,
                 marks=pytest.mark.app(cache='redis://h:6379/1')),
    pytest.param('redis', 'h', 6379, None, 999, None,
                 marks=pytest.mark.app(cache='redis://h:6379/999')),
    pytest.param('redis', 'ho', 6379, 'pw', 0, None,
                 marks=pytest.mark.app(cache='redis://:pw@ho:6379')),
    pytest.param('redis', 'ho', 6379, 'pw', 0, None,
                 marks=pytest.mark.app(cache='redis://user:pw@ho:6379')),
    pytest.param('redis', 'h', 6379, None, 1, {'max_connections': 10},
                 marks=pytest.mark.app(
                     cache='redis://h:6379/1?max_connections=10')),
    pytest.param('redis', 'h', 6, None, 0,
                 {'max_connections': 10, 'stream_timeout': 8},
                 marks=pytest.mark.app(
                     cache='redis://h:6?max_connections=10&stream_timeout=8')),
])
async def test_redis__url(scheme, host, port, password, db, settings,
                          *, app, mocked_redis):
    settings = dict(settings or {})
    settings.setdefault('connect_timeout', None)
    settings.setdefault('stream_timeout', None)
    settings.setdefault('max_connections', None)
    settings.setdefault('max_connections_per_node', None)
    await app.cache.connect()
    mocked_redis.assert_called_once_with(
        host=host,
        port=port,
        password=password,
        db=db,
        skip_full_coverage_check=True,
        **settings)


@pytest.mark.asyncio
@pytest.mark.app(cache='redis://h:6079//')
async def test_redis__url_invalid_path(app, mocked_redis):
    with pytest.raises(ValueError):
        await app.cache.connect()


@pytest.mark.asyncio
@pytest.mark.app(cache='redis://h:6079//')
async def test_redis__using_starting_(app, mocked_redis):
    with pytest.raises(RuntimeError):
        app.cache.client


@pytest.fixture()
def no_aredis(monkeypatch):
    monkeypatch.setattr('faust.web.cache.backends.redis.aredis', None)


@pytest.mark.asyncio
@pytest.mark.app(cache='redis://')
async def test_redis__aredis_is_not_installed(*, app, no_aredis):
    cache = app.cache
    with pytest.raises(ImproperlyConfigured):
        async with cache:
            ...


@pytest.mark.asyncio
@pytest.mark.app(cache='redis://')
async def test_redis__start_twice_same_client(*, app, mocked_redis):
    async with app.cache:
        client1 = app.cache._client
        assert client1 is not None
        client1.ping.assert_called_once_with()
        await app.cache.restart()
        assert app.cache._client is client1


@pytest.mark.asyncio
@pytest.mark.app(cache='redis://')
async def test_redis_get__irrecoverable_errors(*, app, mocked_redis):
    from aredis.exceptions import AuthenticationError
    mocked_redis.return_value.get.coro.side_effect = AuthenticationError()

    with pytest.raises(app.cache.Unavailable):
        async with app.cache:
            await app.cache.get('key')


@pytest.mark.asyncio
@pytest.mark.app(cache='redis://')
@pytest.mark.parametrize('operation,delete_error', [
    ('get', False),
    ('get', True),
    ('delete', False),
    ('delete', True),
])
async def test_redis_invalidating_error(operation, delete_error, *,
                                        app, mocked_redis):
    from aredis.exceptions import DataError
    mocked_op = getattr(mocked_redis.return_value, operation)
    mocked_op.coro.side_effect = DataError()
    if delete_error:
        # then the delete fails
        mocked_redis.return_value.delete.coro.side_effect = DataError()

    with pytest.raises(app.cache.Unavailable):
        async with app.cache:
            await getattr(app.cache, operation)('key')
    if operation == 'delete':
        mocked_redis.return_value.delete.assert_called_with('key')
        assert mocked_redis.return_value.delete.call_count == 2
    else:
        mocked_redis.return_value.delete.assert_called_once_with('key')


@pytest.mark.asyncio
@pytest.mark.app(cache='memory://')
async def test_memory_delete(*, app):
    app.cache.storage.set('foo', b'bar')
    async with app.cache:
        assert await app.cache.get('foo') == b'bar'
        await app.cache.delete('foo')
        assert await app.cache.get('foo') is None


@pytest.mark.asyncio
@pytest.mark.app(cache='redis://')
async def test_redis_get__operational_error(*, app, mocked_redis):
    from aredis.exceptions import TimeoutError
    mocked_redis.return_value.get.coro.side_effect = TimeoutError()

    with pytest.raises(app.cache.Unavailable):
        async with app.cache:
            await app.cache.get('key')


def test_cache_repr(*, app):
    assert repr(app.cache)


async def model_response(response, *,
                         expected_status: int = 200,
                         expected_content_type: str = 'application/json',
                         model=ResponseModel):
    assert response.status == 200
    assert response.content_type == expected_content_type
    return model.from_data(await response.json())


async def model_value(response, **kwargs):
    return (await model_response(response, **kwargs)).value


@pytest.fixture()
def bp(app):
    blueprint.register(app, url_prefix='/test/')


class test_RedisScheme:

    def test_single_client(self, app):
        url = 'redis://123.123.123.123:3636//1'
        Backend = backends.by_url(url)
        assert Backend is redis.CacheBackend
        backend = Backend(app, url=url)
        assert isinstance(backend, redis.CacheBackend)
        client = backend._new_client()
        assert isinstance(client, aredis.StrictRedis)
        pool = client.connection_pool
        assert pool.connection_kwargs['host'] == backend.url.host
        assert pool.connection_kwargs['port'] == backend.url.port
        assert pool.connection_kwargs['db'] == 1

    def test_cluster_client(self, app):
        url = 'rediscluster://123.123.123.123:3636//1'
        Backend = backends.by_url(url)
        assert Backend is redis.CacheBackend
        backend = Backend(app, url=url)
        assert isinstance(backend, redis.CacheBackend)
        client = backend._new_client()
        assert isinstance(client, aredis.StrictRedisCluster)
        pool = client.connection_pool
        assert {'host': backend.url.host,
                'port': 3636} in pool.nodes.startup_nodes



"""Consumer - fetching messages and managing consumer state.

The Consumer is responsible for:

   - Holds reference to the transport that created it

   - ... and the app via ``self.transport.app``.

   - Has a callback that usually points back to ``Conductor.on_message``.

   - Receives messages and calls the callback for every message received.

   - Keeps track of the message and its acked/unacked status.

   - The Conductor forwards the message to all Streams that subscribes
     to the topic the message was sent to.

       + Messages are reference counted, and the Conductor increases
         the reference count to the number of subscribed streams.

       + ``Stream.__aiter__`` is set up in a way such that when what is
         iterating over the stream is finished with the message, a
         finally: block will decrease the reference count by one.

       + When the reference count for a message hits zero, the stream will
         call ``Consumer.ack(message)``, which will mark that topic +
         partition + offset combination as "committable"

       + If all the streams share the same key_type/value_type,
         the conductor will only deserialize the payload once.

   - Commits the offset at an interval

      + The Consumer has a background thread that periodically commits the
        offset.

      - If the consumer marked an offset as committable this thread
        will advance the committed offset.

      + To find the offset that it can safely advance to the commit thread
        will traverse the _acked mapping of TP to list of acked offsets, by
        finding a range of consecutive acked offsets (see note in
        _new_offset).

"""
import abc
import asyncio
import gc
import typing

from collections import defaultdict
from time import monotonic
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSet,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    cast,
)
from weakref import WeakSet
from sortedcontainers import SortedSet

from mode import Service, ServiceT, flight_recorder, get_logger
from mode.threads import MethodQueue, QueueServiceThread
from mode.utils.futures import notify
from mode.utils.locks import Event
from mode.utils.text import pluralize
from mode.utils.times import Seconds
from faust.exceptions import ProducerSendError
from intervaltree import IntervalTree
from faust.types import AppT, ConsumerMessage, Message, RecordMetadata, TP
from faust.types.core import HeadersArg
from faust.types.transports import (
    ConsumerCallback,
    ConsumerT,
    PartitionsAssignedCallback,
    PartitionsRevokedCallback,
    ProducerT,
    TPorTopicSet,
    TransactionManagerT,
    TransportT,
)
from faust.types.tuples import FutureMessage
from faust.utils import terminal
from faust.utils.functional import consecutive_numbers
from faust.utils.tracing import traced_from_parent_span

if typing.TYPE_CHECKING:  # pragma: no cover
    from faust.app import App as _App
else:
    class _App: ...  # noqa: E701

__all__ = ['Consumer', 'Fetcher']

# These flags are used for Service.diag, tracking what the consumer
# service is currently doing.
CONSUMER_FETCHING = 'FETCHING'
CONSUMER_PARTITIONS_REVOKED = 'PARTITIONS_REVOKED'
CONSUMER_PARTITIONS_ASSIGNED = 'PARTITIONS_ASSIGNED'
CONSUMER_COMMITTING = 'COMMITTING'
CONSUMER_SEEKING = 'SEEKING'
CONSUMER_WAIT_EMPTY = 'WAIT_EMPTY'

logger = get_logger(__name__)

RecordMap = Mapping[TP, List[Any]]


class TopicPartitionGroup(NamedTuple):
    """Tuple of ``(topic, partition, group)``."""

    topic: str
    partition: int
    group: int


def ensure_TP(tp: Any) -> TP:
    """Convert aiokafka ``TopicPartition`` to Faust ``TP``."""
    return tp if isinstance(tp, TP) else TP(tp.topic, tp.partition)


def ensure_TPset(tps: Iterable[Any]) -> Set[TP]:
    """Convert set of aiokafka ``TopicPartition`` to Faust ``TP``."""
    return {ensure_TP(tp) for tp in tps}


class Fetcher(Service):
    """Service fetching messages from Kafka."""

    app: AppT

    logger = logger
    _drainer: Optional[asyncio.Future] = None

    def __init__(self, app: AppT, **kwargs: Any) -> None:
        self.app = app
        super().__init__(**kwargs)

    async def on_stop(self) -> None:
        """Call when the fetcher is stopping."""
        if self._drainer is not None and not self._drainer.done():
            self._drainer.cancel()
            while True:
                try:
                    await asyncio.wait_for(self._drainer, timeout=1.0)
                except StopIteration:
                    # Task is cancelled right before coro stops.
                    break
                except asyncio.CancelledError:
                    break
                except asyncio.TimeoutError:
                    self.log.warning('Fetcher is ignoring cancel or slow :(')
                else:  # pragma: no cover
                    # coverage does not record this line as being executed
                    # but I've verified that it is [ask]
                    break

    @Service.task
    async def _fetcher(self) -> None:
        try:
            consumer = cast(Consumer, self.app.consumer)
            self._drainer = asyncio.ensure_future(
                consumer._drain_messages(self),
                loop=self.loop,
            )
            await self._drainer
        except asyncio.CancelledError:
            pass
        finally:
            self.set_shutdown()


class TransactionManager(Service, TransactionManagerT):
    """Manage producer transactions."""

    app: AppT

    transactional_id_format = '{group_id}-{tpg.group}-{tpg.partition}'

    def __init__(self, transport: TransportT,
                 *,
                 consumer: 'ConsumerT',
                 producer: 'ProducerT',
                 **kwargs: Any) -> None:
        self.transport = transport
        self.app = self.transport.app
        self.consumer = consumer
        self.producer = producer
        super().__init__(**kwargs)

    async def on_stop(self) -> None:
        """Call when transaction manager is stopping."""
        # Final flush to ensure no pending data
        await self.flush()

        self.log.info('Stopping all active transactions due to application shutdown')
        if self.consumer.assignment():
            # Get the transaction IDs from all currently assigned partitions
            transaction_ids = sorted(self._tps_to_all_transactional_ids(self.consumer.assignment()))
            if transaction_ids:
                self.log.info(
                    'Stopping %r transactional %s...',
                    len(transaction_ids),
                    pluralize(len(transaction_ids), 'producer')
                )
                await self._stop_transactions(transaction_ids)

    async def flush(self) -> None:
        """Wait for producer to transmit all pending messages."""
        await self.producer.flush()

    async def on_partitions_revoked(self, revoked: Set[TP]) -> None:
        """Call when the cluster is rebalancing and partitions are revoked."""
        await traced_from_parent_span()(self.flush)()

    async def on_rebalance(self,
                           assigned: Set[TP],
                           revoked: Set[TP],
                           newly_assigned: Set[TP]) -> None:
        """Call when the cluster is rebalancing."""
        T = traced_from_parent_span()
        
        # First, ensure we're done with any pending transactions
        await T(self.flush)()
        
        # Stop producers for revoked partitions - use _tps_to_all_transactional_ids to 
        # include ALL revoked partitions, both active and standby
        revoked_tids = sorted(self._tps_to_all_transactional_ids(revoked))
        if revoked_tids:
            self.log.info(
                'Stopping %r transactional %s for %r revoked %s...',
                len(revoked_tids),
                pluralize(len(revoked_tids), 'producer'),
                len(revoked),
                pluralize(len(revoked), 'partition'))
            await T(self._stop_transactions, tids=revoked_tids)(revoked_tids)
            
        # Give the broker some time to clean up the old transactions
        await asyncio.sleep(0.1)

        # Start producers for assigned partitions
        # Only start transactions for newly assigned partitions to avoid conflicts
        # when scaling down and reusing the same transactional IDs
        assigned_tids = sorted(self._tps_to_transactional_ids(assigned))
        if assigned_tids:
            self.log.info(
                'Starting %r transactional %s for %r assigned %s...',
                len(assigned_tids),
                pluralize(len(assigned_tids), 'producer'),
                len(assigned),
                pluralize(len(assigned), 'partition'))
            await T(self._start_transactions,
                    tids=assigned_tids)(assigned_tids)

    async def _stop_transactions(self, tids: Iterable[str]) -> None:
        T = traced_from_parent_span()
        producer = self.producer
        for transactional_id in tids:
            await T(producer.stop_transaction)(transactional_id)

    async def _start_transactions(self, tids: Iterable[str]) -> None:
        T = traced_from_parent_span()
        producer = self.producer
        for transactional_id in tids:
            await T(producer.maybe_begin_transaction)(transactional_id)

    def _tps_to_transactional_ids(self, tps: Set[TP]) -> Set[str]:
        return {
            self.transactional_id_format.format(
                tpg=tpg,
                group_id=self.app.conf.id,
            )
            for tpg in self._tps_to_active_tpgs(tps)
        }

    def _tps_to_all_tpgs(self, tps: Set[TP]) -> Set[TopicPartitionGroup]:
        """Convert all topic partitions (including standbys) to TopicPartitionGroups.
        
        This is used for revoked partitions where we need to stop ALL transactions,
        regardless of whether they were active or standby.
        """
        assignor = self.app.assignor
        return {
            TopicPartitionGroup(
                tp.topic,
                tp.partition,
                assignor.group_for_topic(tp.topic),
            )
            for tp in tps
        }

    def _tps_to_all_transactional_ids(self, tps: Set[TP]) -> Set[str]:
        """Get transaction IDs for all partitions, including standbys.
        
        Used for cleanup during partition revocation.
        """
        return {
            self.transactional_id_format.format(
                tpg=tpg,
                group_id=self.app.conf.id,
            )
            for tpg in self._tps_to_all_tpgs(tps)
        }
    
    def _tps_to_active_tpgs(self, tps: Set[TP]) -> Set[TopicPartitionGroup]:
        assignor = self.app.assignor
        return {
            TopicPartitionGroup(
                tp.topic,
                tp.partition,
                assignor.group_for_topic(tp.topic),
            )
            for tp in tps
            if not assignor.is_standby(tp)
        }

    async def send(self, topic: str, key: Optional[bytes],
                   value: Optional[bytes],
                   partition: Optional[int],
                   timestamp: Optional[float],
                   headers: Optional[HeadersArg],
                   *,
                   transactional_id: str = None) -> Awaitable[RecordMetadata]:
        """Schedule message to be sent by producer."""
        group = transactional_id = None
        p = self.consumer.key_partition(topic, key, partition)
        if p is not None:
            group = self.app.assignor.group_for_topic(topic)
            transactional_id = f'{self.app.conf.id}-{group}-{p}'
        elif partition is not None:
            p = partition            
        return await self.producer.send(
            topic, key, value, p, timestamp, headers,
            transactional_id=transactional_id,
        )

    def send_soon(self, fut: FutureMessage) -> None:
        raise NotImplementedError()

    async def send_and_wait(self, topic: str, key: Optional[bytes],
                            value: Optional[bytes],
                            partition: Optional[int],
                            timestamp: Optional[float],
                            headers: Optional[HeadersArg],
                            *,
                            transactional_id: str = None) -> RecordMetadata:
        """Send message and wait for it to be transmitted."""
        fut = await self.send(topic, key, value, partition, timestamp, headers)
        return await fut

    async def commit(self, offsets: Mapping[TP, int],
                     start_new_transaction: bool = True) -> bool:
        """Commit offsets for partitions."""
        producer = self.producer
        group_id = self.app.conf.id
        by_transactional_id: MutableMapping[str, MutableMapping[TP, int]]
        by_transactional_id = defaultdict(dict)

        for tp, offset in offsets.items():
            group = self.app.assignor.group_for_topic(tp.topic)
            transactional_id = f'{group_id}-{group}-{tp.partition}'
            by_transactional_id[transactional_id][tp] = offset

        if by_transactional_id:
            await producer.commit_transactions(
                by_transactional_id, group_id,
                start_new_transaction=start_new_transaction,
            )
        return True

    def key_partition(self, topic: str, key: bytes) -> TP:
        raise NotImplementedError()

    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 30.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        """Create/declare topic on server."""
        return await self.producer.create_topic(
            topic, partitions, replication,
            config=config,
            timeout=timeout,
            retention=retention,
            compacting=compacting,
            deleting=deleting,
            ensure_created=ensure_created,
        )

    def supports_headers(self) -> bool:
        """Return :const:`True` if the Kafka server supports headers."""
        return self.producer.supports_headers()


class Consumer(Service, ConsumerT):
    """Base Consumer."""

    app: AppT

    logger = logger

    #: Tuple of exception types that may be raised when the
    #: underlying consumer driver is stopped.
    consumer_stopped_errors: ClassVar[Tuple[Type[BaseException], ...]] = ()

    # Mapping of TP to list of gap in offsets.
    _gap: MutableMapping[TP, List[int]]

    # Mapping of TP to sorted set of acked offsets.
    _acked: MutableMapping[TP, 'SortedSet[int]']

    #: Fast lookup to see if tp+offset was acked.
    _acked_index: MutableMapping[TP, Set[int]]

    #: Keeps track of the currently read offset in each TP
    _read_offset: MutableMapping[TP, Optional[int]]

    #: Keeps track of the currently committed offset in each TP.
    _committed_offset: MutableMapping[TP, Optional[int]]

    #: The consumer.wait_empty() method will set this to be notified
    #: when something acks a message.
    _waiting_for_ack: Optional[asyncio.Future] = None

    #: Used by .commit to ensure only one thread is comitting at a time.
    #: Other thread starting to commit while a commit is already active,
    #: will wait for the original request to finish, and do nothing.
    _commit_fut: Optional[asyncio.Future] = None

    #: Set of unacked messages: that is messages that we started processing
    #: and that we MUST attempt to complete processing of, before
    #: shutting down or resuming a rebalance.
    _unacked_messages: MutableSet[Message]

    #: Time of when the consumer was started.
    _time_start: float

    # How often to poll and track log end offsets.
    _end_offset_monitor_interval: float

    _commit_every: Optional[int]
    _n_acked: int = 0

    _active_partitions: Optional[Set[TP]]
    _paused_partitions: Set[TP]

    flow_active: bool = True
    can_resume_flow: Event

    #: Ensures consumer isn't in middle of writing
    #: records into queue during rebalance
    can_stop_flow: Event
    
    def __init__(self,
                 transport: TransportT,
                 callback: ConsumerCallback,
                 on_partitions_revoked: PartitionsRevokedCallback,
                 on_partitions_assigned: PartitionsAssignedCallback,
                 *,
                 commit_interval: float = None,
                 commit_livelock_soft_timeout: float = None,
                 loop: asyncio.AbstractEventLoop = None,
                 **kwargs: Any) -> None:
        assert callback is not None
        self.transport = transport
        self.app = self.transport.app
        self.in_transaction = self.app.in_transaction
        self.callback = callback
        self._on_message_in = self.app.sensors.on_message_in
        self._on_partitions_revoked = on_partitions_revoked
        self._on_partitions_assigned = on_partitions_assigned
        self._commit_every = self.app.conf.broker_commit_every
        self.scheduler = self.app.conf.ConsumerScheduler()
        self.commit_interval = (
            commit_interval or self.app.conf.broker_commit_interval)
        self.commit_livelock_soft_timeout = (
            commit_livelock_soft_timeout or
            self.app.conf.broker_commit_livelock_soft_timeout)
        self._gap = defaultdict(IntervalTree)
        self._acked = defaultdict(SortedSet)
        self._acked_index = defaultdict(set)
        self._read_offset = defaultdict(lambda: None)
        self._committed_offset = defaultdict(lambda: None)
        self._unacked_messages = WeakSet()
        self._waiting_for_ack = None
        self._time_start = monotonic()
        self._end_offset_monitor_interval = self.commit_interval * 2
        self.randomly_assigned_topics = set()
        self.can_resume_flow = Event()
        self.can_stop_flow = Event()
        self._reset_state()
        super().__init__(loop=loop or self.transport.loop, **kwargs)
        self.transactions = self.transport.create_transaction_manager(
            consumer=self,
            producer=self.app.producer,
            beacon=self.beacon,
            loop=self.loop,
        )

    def on_init_dependencies(self) -> Iterable[ServiceT]:
        """Return list of services this consumer depends on."""
        # We start the TransactionManager only if
        # processing_guarantee='exactly_once'
        if self.in_transaction:
            return [self.transactions]
        return []

    def _reset_state(self) -> None:
        self._active_partitions = None
        self._paused_partitions = set()
        self.can_resume_flow.clear()
        self.can_stop_flow.clear()
        self.flow_active = True
        self._time_start = monotonic()

    async def on_restart(self) -> None:
        """Call when the consumer is restarted."""
        self._reset_state()
        self.on_init()

    def _get_active_partitions(self) -> Set[TP]:
        tps = self._active_partitions
        if tps is None:
            return self._set_active_tps(self.assignment())
        assert all(isinstance(x, TP) for x in tps)
        return tps

    def _set_active_tps(self, tps: Set[TP]) -> Set[TP]:
        xtps = self._active_partitions = ensure_TPset(tps)  # copy
        xtps.difference_update(self._paused_partitions)
        return xtps

    @abc.abstractmethod
    async def _commit(
            self,
            offsets: Mapping[TP, int]) -> bool:  # pragma: no cover
        ...

    async def perform_seek(self) -> None:
        """Seek all partitions to their current committed position."""
        read_offset = self._read_offset
        _committed_offsets = await self.seek_to_committed()
        read_offset.update({
            # Since committed offset is "next to consume", 
            # read_offset should be the last processed (committed - 1)
            tp: (offset - 1) if offset is not None and offset > 0 else None
            for tp, offset in _committed_offsets.items()
        })
        committed_offsets = {
            ensure_TP(tp): offset if offset else None
            for tp, offset in _committed_offsets.items()
            if offset is not None
        }
        self._committed_offset.update(committed_offsets)

    @abc.abstractmethod
    async def seek_to_committed(self) -> Mapping[TP, int]:
        """Seek all partitions to their committed offsets."""
        ...

    async def seek(self, partition: TP, offset: int) -> None:
        """Seek partition to specific offset."""
        self.log.dev('SEEK %r -> %r', partition, offset)
        # reset livelock detection
        await self._seek(partition, offset)
        # set new read offset so we will reread messages
        self._read_offset[ensure_TP(partition)] = offset if offset else None

    @abc.abstractmethod
    async def _seek(self, partition: TP, offset: int) -> None:
        ...

    async def stop_flow(self) -> None:
        """Block consumer from processing any more messages."""
        self.flow_active = False
        self.can_resume_flow.clear()
        await self.wait(self.can_stop_flow, timeout=5)

    def resume_flow(self) -> None:
        """Allow consumer to process messages."""
        self.flow_active = True
        self.can_resume_flow.set()
        self.can_stop_flow.clear()

    def pause_partitions(self, tps: Iterable[TP]) -> None:
        """Pause fetching from partitions."""
        tpset = ensure_TPset(tps)
        self._get_active_partitions().difference_update(tpset)
        self._paused_partitions.update(tpset)

    def resume_partitions(self, tps: Iterable[TP]) -> None:
        """Resume fetching from partitions."""
        tpset = ensure_TPset(tps)
        self._get_active_partitions().update(tps)
        self._paused_partitions.difference_update(tpset)

    @abc.abstractmethod
    def _new_topicpartition(
            self, topic: str, partition: int) -> TP:  # pragma: no cover
        ...

    def _is_changelog_tp(self, tp: TP) -> bool:
        return tp.topic in self.app.tables.changelog_topics

    @Service.transitions_to(CONSUMER_PARTITIONS_REVOKED)
    async def on_partitions_revoked(self, revoked: Set[TP]) -> None:
        """Call during rebalancing when partitions are being revoked."""
        # NOTE:
        # The ConsumerRebalanceListener is responsible for calling
        # app.on_rebalance_start(), and this must have happened
        # before we get to this point (see aiokafka implementation).
        span = self.app._start_span_from_rebalancing('on_partitions_revoked')
        T = traced_from_parent_span(span)
        with span:
            # see comment in on_partitions_assigned
            # remove revoked partitions from active + paused tps.
            if self._active_partitions is not None:
                self._active_partitions.difference_update(revoked)
            self._paused_partitions.difference_update(revoked)
            # Remove the revoked partitions from local data structures
            for tp in revoked:
                self._gap.pop(tp, None)
                self._acked.pop(tp, None)
                self._acked_index.pop(tp, None)
                self._read_offset.pop(tp, None)
                self._committed_offset.pop(tp, None)            
            await T(self._on_partitions_revoked, partitions=revoked)(revoked)
            

    @Service.transitions_to(CONSUMER_PARTITIONS_ASSIGNED)
    async def on_partitions_assigned(self, assigned: Set[TP]) -> None:
        """Call during rebalancing when partitions are being assigned."""
        span = self.app._start_span_from_rebalancing('on_partitions_assigned')
        T = traced_from_parent_span(span)
        with span:
            # remove recently revoked tps from set of paused tps.
            self._paused_partitions.intersection_update(assigned)
            # cache set of assigned partitions
            self._set_active_tps(assigned)
            # start callback chain of assigned callbacks.
            #   need to copy set at this point, since we cannot have
            #   the callbacks mutate our active list.
            await T(self._on_partitions_assigned, partitions=assigned)(
                assigned)
        self.app.on_rebalance_return()

    @abc.abstractmethod
    async def _getmany(self,
                       active_partitions: Optional[Set[TP]],
                       timeout: float) -> RecordMap:
        ...

    async def getmany(self,
                      timeout: float) -> AsyncIterator[Tuple[TP, Message]]:
        """Fetch batch of messages from server."""
        # records' contain mapping from TP to list of messages.
        # if there are two agents, consuming from topics t1 and t2,
        # normal order of iteration would be to process each
        # tp in the dict:
        #    for tp. messages in records.items():
        #        for message in messages:
        #           yield tp, message
        #
        # The problem with this, is if we have prefetched 16k records
        # for one partition, the other partitions won't even start processing
        # before those 16k records are completed.
        #
        # So we try round-robin between the tps instead:
        #
        #    iterators: Dict[TP, Iterator] = {
        #        tp: iter(messages)
        #        for tp, messages in records.items()
        #    }
        #    while iterators:
        #        for tp, messages in iterators.items():
        #            yield tp, next(messages)
        #            # remove from iterators if empty.
        #
        # The problem with this implementation is that
        # the records mapping is ordered by TP, so records.keys()
        # will look like this:
        #
        #  TP(topic='bar', partition=0)
        #  TP(topic='bar', partition=1)
        #  TP(topic='bar', partition=2)
        #  TP(topic='bar', partition=3)
        #  TP(topic='foo', partition=0)
        #  TP(topic='foo', partition=1)
        #  TP(topic='foo', partition=2)
        #  TP(topic='foo', partition=3)
        #
        # If there are 100 partitions for each topic,
        # it will process 100 items in the first topic, then 100 items
        # in the other topic, but even worse if partition counts
        # vary greatly, t1 has 1000 partitions and t2
        # has 1 partition, then t2 will end up being starved most of the time.
        #
        # We solve this by going round-robin through each topic.
        records, active_partitions = await self._wait_next_records(timeout)
        if records is None or self.should_stop:
            return

        records_it = self.scheduler.iterate(records)
        to_message = self._to_message  # localize
        if self.flow_active:
            for tp, record in records_it:
                if not self.flow_active:
                    break
                if active_partitions is None or tp in active_partitions:
                    highwater_mark = self.highwater(tp)
                    self.app.monitor.track_tp_end_offset(tp, highwater_mark)
                    # convert timestamp to seconds from int milliseconds.
                    yield tp, to_message(tp, record)

    async def _wait_next_records(
            self, timeout: float) -> Tuple[Optional[RecordMap],
                                           Optional[Set[TP]]]:
        if not self.flow_active:
            self.log.dev('Flow is inactive. Waiting to resume...')
            self.can_stop_flow.set()            
            await self.wait(self.can_resume_flow)
        # Implementation for the Fetcher service.

        is_client_only = self.app.client_only

        active_partitions: Optional[Set[TP]]
        if is_client_only:
            active_partitions = None
        else:
            active_partitions = self._get_active_partitions()

        records: RecordMap = {}
        if is_client_only or active_partitions:
            # Fetch records only if active partitions to avoid the risk of
            # fetching all partitions in the beginning when none of the
            # partitions is paused/resumed.
            records = await self._getmany(
                active_partitions=active_partitions,
                timeout=timeout,
            )
        else:
            # We should still release to the event loop
            await self.sleep(1)
        return records, active_partitions

    @abc.abstractmethod
    def _to_message(self, tp: TP, record: Any) -> ConsumerMessage:
        ...

    def track_message(self, message: Message) -> None:
        """Track message and mark it as pending ack."""
        # add to set of pending messages that must be acked for graceful
        # shutdown.  This is called by transport.Conductor,
        # before delivering messages to streams.
        self._unacked_messages.add(message)
        # call sensors
        self._on_message_in(message.tp, message.offset, message)

    def ack(self, message: Message) -> bool:
        """Mark message as being acknowledged by stream."""
        if not message.acked:
            message.acked = True
            tp = message.tp
            offset = message.offset
            if self.app.topics.acks_enabled_for(message.topic):
                committed = self._committed_offset[tp]
                try:
                    if committed is None or offset >= committed:
                        acked_index = self._acked_index[tp]
                        if offset not in acked_index:
                            self._unacked_messages.discard(message)
                            acked_index.add(offset)
                            acked_for_tp = self._acked[tp]
                            acked_for_tp.add(offset)
                            self._n_acked += 1
                            return True
                finally:
                    notify(self._waiting_for_ack)
        return False

    async def _wait_for_ack(self, timeout: float) -> None:
        # arm future so that `ack()` can wake us up
        self._waiting_for_ack = asyncio.Future(loop=self.loop)
        try:
            # wait for `ack()` to wake us up
            await asyncio.wait_for(
                self._waiting_for_ack, loop=self.loop, timeout=1)
        except (asyncio.TimeoutError,
                asyncio.CancelledError):  # pragma: no cover
            pass
        finally:
            self._waiting_for_ack = None

    @Service.transitions_to(CONSUMER_WAIT_EMPTY)
    async def wait_empty(self) -> None:
        """Wait for all messages that started processing to be acked."""
        wait_count = 0
        T = traced_from_parent_span()
        while not self.should_stop and self._unacked_messages:
            wait_count += 1
            if not wait_count % 10:  # pragma: no cover
                remaining = [(m.refcount, m) for m in self._unacked_messages]
                self.log.warning('wait_empty: Waiting for tasks %r', remaining)
                self.log.info(
                    'Agent tracebacks:\n%s',
                    self.app.agents.human_tracebacks(),
                )
            self.log.dev('STILL WAITING FOR ALL STREAMS TO FINISH')
            self.log.dev('WAITING FOR %r EVENTS', len(self._unacked_messages))
            gc.collect()
            await T(self.commit)()
            if not self._unacked_messages:
                break
            await T(self._wait_for_ack)(timeout=1)
            self._clean_unacked_messages()

        self.log.dev('COMMITTING AGAIN AFTER STREAMS DONE')
        await T(self.commit_and_end_transactions)()

    def _clean_unacked_messages(self) -> None:
        # remove actually acked messages from weakset.
        self._unacked_messages -= {
            message for message in self._unacked_messages
            if message.acked
        }

    async def commit_and_end_transactions(self) -> None:
        """Commit all safe offsets and end transaction."""
        await self.commit(start_new_transaction=False)

    async def on_stop(self) -> None:
        """Call when consumer is stopping."""
        if self.app.conf.stream_wait_empty:
            await self.wait_empty()
        else:
            await self.commit_and_end_transactions()

    @Service.task
    async def _commit_handler(self) -> None:
        interval = self.commit_interval

        await self.sleep(interval)
        async for sleep_time in self.itertimer(interval, name='commit'):
            await self.commit()

    @Service.task
    async def _commit_livelock_detector(self) -> None:  # pragma: no cover
        interval: float = self.commit_interval * 2.5
        await self.sleep(interval)
        async for sleep_time in self.itertimer(interval, name='livelock'):
            if not self.app.rebalancing:
                await self.verify_all_partitions_active()

    async def verify_all_partitions_active(self) -> None:
        now = monotonic()
        for tp in self.assignment():
            await self.sleep(0)
            if not self.should_stop:
                self.verify_event_path(now, tp)

    def verify_event_path(self, now: float, tp: TP) -> None:
        ...

    def verify_recovery_event_path(self, now: float, tp: TP) -> None:
        ...

    async def commit(self, topics: TPorTopicSet = None,
                     start_new_transaction: bool = True) -> bool:
        """Maybe commit the offset for all or specific topics.

        Arguments:
            topics: Set containing topics and/or TopicPartitions to commit.
        """
        if self.app.client_only:
            # client only cannot commit as consumer does not have group_id
            return False
        if await self.maybe_wait_for_commit_to_finish():
            # original commit finished, return False as we did not commit
            return False

        self._commit_fut = asyncio.Future(loop=self.loop)
        try:
            return await self.force_commit(
                topics,
                start_new_transaction=start_new_transaction,
            )
        finally:
            # set commit_fut to None so that next call will commit.
            fut, self._commit_fut = self._commit_fut, None
            # notify followers that the commit is done.
            notify(fut)

    async def maybe_wait_for_commit_to_finish(self) -> bool:
        """Wait for any existing commit operation to finish."""
        # Only one coroutine allowed to commit at a time,
        # and other coroutines should wait for the original commit to finish
        # then do nothing.
        if self._commit_fut is not None:
            # something is already committing so wait for that future.
            try:
                await self._commit_fut
            except asyncio.CancelledError:
                # if future is cancelled we have to start new commit
                pass
            else:
                return True
        return False

    @Service.transitions_to(CONSUMER_COMMITTING)
    async def force_commit(self,
                           topics: TPorTopicSet = None,
                           start_new_transaction: bool = True) -> bool:
        """Force offset commit."""
        sensor_state = self.app.sensors.on_commit_initiated(self)

        # Go over the ack list in each topic/partition
        commit_tps = list(self._filter_tps_with_pending_acks(topics))
        did_commit = await self._commit_tps(
            commit_tps, start_new_transaction=start_new_transaction)

        self.app.sensors.on_commit_completed(self, sensor_state)
        return did_commit

    async def _commit_tps(self,
                          tps: Iterable[TP],
                          start_new_transaction: bool) -> bool:
        commit_offsets = self._filter_committable_offsets(tps)
        if commit_offsets:
            try:
                # send all messages attached to the new offset
                await self._handle_attached(commit_offsets)
            except ProducerSendError as exc:
                await self.crash(exc)
            else:
                return await self._commit_offsets(
                    commit_offsets,
                    start_new_transaction=start_new_transaction)
        return False

    def _filter_committable_offsets(self, tps: Iterable[TP]) -> Dict[TP, int]:
        commit_offsets = {}
        for tp in tps:
            # Find the latest offset we can commit in this tp
            offset = self._new_offset(tp)
            # check if we can commit to this offset
            if offset is not None and self._should_commit(tp, offset):
                commit_offsets[tp] = offset
        return commit_offsets

    async def _handle_attached(self, commit_offsets: Mapping[TP, int]) -> None:
        for tp, offset in commit_offsets.items():
            app = cast(_App, self.app)
            attachments = app._attachments
            producer = app.producer
            # Start publishing the messages and return a list of pending
            # futures.
            pending = await attachments.publish_for_tp_offset(tp, offset)
            # then we wait for either
            #  1) all the attached messages to be published, or
            #  2) the producer crashing
            #
            # If the producer crashes we will not be able to send any messages
            # and it only crashes when there's an irrecoverable error.
            #
            # If we cannot commit it means the events will be processed again,
            # so conforms to at-least-once semantics.
            if pending:
                await cast(Service, producer).wait_many(pending)

    async def _commit_offsets(self, offsets: Mapping[TP, int],
                              start_new_transaction: bool = True) -> bool:
        table = terminal.logtable(
            [(str(tp), str(offset))
             for tp, offset in offsets.items()],
            title='Commit Offsets',
            headers=['TP', 'Offset'],
        )
        self.log.dev('COMMITTING OFFSETS:\n%s', table)
        assignment = self.assignment()
        committable_offsets: Dict[TP, int] = {}
        revoked: Dict[TP, int] = {}
        for tp, offset in offsets.items():
            if tp in assignment:
                committable_offsets[tp] = offset
            else:
                revoked[tp] = offset
        if revoked:
            self.log.info(
                'Discarded commit for revoked partitions that '
                'will be eventually processed again: %r',
                revoked,
            )
        if not committable_offsets:
            return False
        with flight_recorder(self.log, timeout=300.0) as on_timeout:
            did_commit = False
            on_timeout.info('+consumer.commit()')
            if self.in_transaction:
                did_commit = await self.transactions.commit(
                    committable_offsets,
                    start_new_transaction=start_new_transaction,
                )
            else:
                did_commit = await self._commit(committable_offsets)
            on_timeout.info('-consumer.commit()')
            if did_commit:
                on_timeout.info('+tables.on_commit')
                self.app.tables.on_commit(committable_offsets)
                on_timeout.info('-tables.on_commit')
        self._committed_offset.update(committable_offsets)
        self.app.monitor.on_tp_commit(committable_offsets)
        return did_commit

    def _filter_tps_with_pending_acks(
            self, topics: TPorTopicSet = None) -> Iterator[TP]:
        return (tp for tp in self._acked
                if topics is None or tp in topics or tp.topic in topics)

    def _should_commit(self, tp: TP, offset: int) -> bool:
        committed = self._committed_offset[tp]
        return committed is None or bool(offset) and offset > committed

    def _new_offset(self, tp: TP) -> Optional[int]:
        # get the new offset for this tp, by going through
        # its sorted set of acked messages.
        acked = self._acked[tp]

        # We iterate over it until we find a gap
        # then return the offset before that.
        # For example if acked[tp] is:
        #   1 2 3 4 5 6 7 8 9
        # the return value will be: 9
        # If acked[tp] is:
        #  34 35 36 40 41 42 43 44
        #          ^--- gap
        # the return value will be: 36
        if acked:
            max_offset = max(acked)
            gap_for_tp: IntervalTree = self._gap[tp]
            if gap_for_tp:
                # find all the ranges up to the max of acked, add them in to acked,
                # and chop them off the gap.
                candidates = gap_for_tp.overlap(0, max_offset)
                # note: merge_overlaps will sort the intervaltree and will ensure that
                # the intervals left over don't overlap each other. So can sort by their
                # start without worrying about ends overlapping.
                sorted_candidates = sorted(candidates, key=lambda x: x.begin)
                if sorted_candidates:
                    stuff_to_add = []
                    for entry in sorted_candidates:
                        stuff_to_add.extend(range(entry.begin, entry.end))
                    new_max_offset = max(stuff_to_add[-1], max_offset + 1)
                    # Add all items to the sorted set
                    acked.update(stuff_to_add)
                    gap_for_tp.chop(0, new_max_offset)

            # We iterate over it until we handle gap in the head of acked queue
            # then return the previous committed offset.
            # For example if acked[tp] is:
            #    34 35 36 37
            #  ^-- gap
            # self._committed_offset[tp] is 32 (next offset to consume after 31)
            # the return value will be None (the same as 31)
            if self._committed_offset[tp]:
                if min(acked) - self._committed_offset[tp] > 0:
                    return None

            # Note: acked is always kept sorted because it's a SortedSet.
            # find first list of consecutive numbers
            batch = next(consecutive_numbers(acked))
            
            # Remove batch items from the sorted set
            to_remove = set(batch[:-1])  # All but the last item
            if to_remove:
                acked.difference_update(to_remove)
            self._acked_index[tp].difference_update(batch)
            
            # return the next offset to consume (last processed + 1)
            # This prevents duplicate consumption on restart since Kafka
            # will start delivering from this offset (exclusive of processed messages)
            return batch[-1] + 1
        return None

    async def on_task_error(self, exc: BaseException) -> None:
        """Call when processing a message failed."""
        await self.commit()

    async def _add_gap(self, tp: TP, offset_from: int, offset_to: int) -> None:
        committed = self._committed_offset[tp]
        gap_for_tp = self._gap[tp]
        if committed is not None:
            # committed is "next offset to consume", so last processed is committed - 1
            offset_from = max(offset_from, committed)
        # intervaltree intervals exclude the end
        if offset_from <= offset_to:
            gap_for_tp.addi(offset_from, offset_to + 1)
            # sleep 0 to allow other coroutines to get some loop time
            # for example, to answer health checks while building the gap
            await asyncio.sleep(0)
            gap_for_tp.merge_overlaps()

    async def _drain_messages(self, fetcher: ServiceT) -> None:  # pragma: no cover
        # This is the background thread started by Fetcher, used to
        # constantly read messages using Consumer.getmany.
        # It takes Fetcher as argument, because we must be able to
        # stop it using `await Fetcher.stop()`.
        callback = self.callback
        getmany = self.getmany
        consumer_should_stop = cast(Service, self)._stopped.is_set
        fetcher_should_stop = cast(Service, fetcher)._stopped.is_set

        get_read_offset = self._read_offset.__getitem__
        set_read_offset = self._read_offset.__setitem__
        flag_consumer_fetching = CONSUMER_FETCHING
        set_flag = self.diag.set_flag
        unset_flag = self.diag.unset_flag
        commit_every = self._commit_every
        acks_enabled_for = self.app.topics.acks_enabled_for

        yield_every = 100
        num_since_yield = 0
        sleep = asyncio.sleep

        try:
            while not (consumer_should_stop() or fetcher_should_stop()):
                set_flag(flag_consumer_fetching)
                ait = cast(AsyncIterator, getmany(timeout=1.0))

                # Sleeping because sometimes getmany is called in a loop
                # never releasing to the event loop
                await self.sleep(0)
                if not self.should_stop:
                    async for tp, message in ait:
                        num_since_yield += 1
                        if num_since_yield > yield_every:
                            await sleep(0)
                            num_since_yield = 0

                        offset = message.offset
                        r_offset = get_read_offset(tp)
                        if r_offset is None or offset >= r_offset:
                            gap = offset - (r_offset or 0)
                            # We have a gap in income messages
                            # Note: tracking gaps is slow but important for offset commit management.
                            # Changelog topics can have large and frequent gaps due to compaction, but we
                            # don't commit offsets to kafka for changelog topics so we can skip gap tracking for them.
                            # and improve performance for table recovery procedure for highly compacted changelog topics.
                            if gap > 1 and r_offset and not self._is_changelog_tp(tp):
                                acks_enabled = acks_enabled_for(message.topic)
                                if acks_enabled:
                                    await self._add_gap(tp, r_offset + 1, offset)
                            if commit_every is not None:
                                if self._n_acked >= commit_every:
                                    self._n_acked = 0
                                    await self.commit()
                            await callback(message)
                            set_read_offset(tp, offset)
                        else:
                            self.log.dev(
                                "DROPPED MESSAGE ROFF %r: k=%r v=%r",
                                offset,
                                message.key,
                                message.value,
                            )
                    unset_flag(flag_consumer_fetching)

        except self.consumer_stopped_errors:
            if self.transport.app.should_stop:
                # we're already stopping so ignore
                self.log.info("Broker stopped consumer, shutting down...")
                return
            raise
        except asyncio.CancelledError:
            if self.transport.app.should_stop:
                # we're already stopping so ignore
                self.log.info("Consumer shutting down for user cancel.")
                return
            raise
        except Exception as exc:
            self.log.exception("Drain messages raised: %r", exc)
            raise
        finally:
            self.can_stop_flow.set()
            unset_flag(flag_consumer_fetching)

    def close(self) -> None:
        """Close consumer for graceful shutdown."""
        ...

    @property
    def unacked(self) -> Set[Message]:
        """Return the set of currently unacknowledged messages."""
        return cast(Set[Message], self._unacked_messages)


class ConsumerThread(QueueServiceThread):
    """Consumer running in a dedicated thread."""

    app: AppT
    consumer: 'ThreadDelegateConsumer'
    transport: TransportT

    def __init__(self, consumer: 'ThreadDelegateConsumer',
                 **kwargs: Any) -> None:
        self.consumer = consumer
        self.transport = self.consumer.transport
        self.app = self.transport.app
        super().__init__(**kwargs)

    @abc.abstractmethod
    async def subscribe(self, topics: Iterable[str]) -> None:
        """Reset subscription (requires rebalance)."""
        ...

    @abc.abstractmethod
    async def seek_to_committed(self) -> Mapping[TP, int]:
        """Seek all partitions to their committed offsets."""
        ...

    @abc.abstractmethod
    async def commit(self, tps: Mapping[TP, int]) -> bool:
        """Commit offsets in topic partitions."""
        ...

    @abc.abstractmethod
    async def position(self, tp: TP) -> Optional[int]:
        """Return the current offset for partition."""
        ...

    @abc.abstractmethod
    async def seek_to_beginning(self, *partitions: TP) -> None:
        """Seek to the earliest offsets available for partitions."""
        ...

    @abc.abstractmethod
    async def seek_wait(self, partitions: Mapping[TP, int]) -> None:
        """Seek partitions to specific offsets and wait."""
        ...

    @abc.abstractmethod
    def seek(self, partition: TP, offset: int) -> None:
        """Seek partition to specific offset."""
        ...

    @abc.abstractmethod
    def assignment(self) -> Set[TP]:
        """Return the current assignment."""
        ...

    @abc.abstractmethod
    def highwater(self, tp: TP) -> int:
        """Return the last available offset in partition."""
        ...

    @abc.abstractmethod
    def topic_partitions(self, topic: str) -> Optional[int]:
        """Return number of configured partitions for topic by name."""
        ...

    @abc.abstractmethod
    def close(self) -> None:
        ...

    @abc.abstractmethod
    async def earliest_offsets(self, *partitions: TP) -> Mapping[TP, int]:
        """Return the earliest available offset for list of partitions."""
        ...

    @abc.abstractmethod
    async def highwaters(self, *partitions: TP) -> Mapping[TP, int]:
        """Return the last available offset for list of partitions."""
        ...

    @abc.abstractmethod
    async def getmany(self,
                      active_partitions: Optional[Set[TP]],
                      timeout: float) -> RecordMap:
        """Fetch batch of messages from server."""
        ...

    @abc.abstractmethod
    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 30.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        """Create/declare topic on server."""
        ...

    async def on_partitions_revoked(
            self, revoked: Set[TP]) -> None:
        """Call on rebalance when partitions are being revoked."""
        await self.consumer.threadsafe_partitions_revoked(
            self.thread_loop, revoked)

    async def on_partitions_assigned(
            self, assigned: Set[TP]) -> None:
        """Call on rebalance when partitions are being assigned."""
        await self.consumer.threadsafe_partitions_assigned(
            self.thread_loop, assigned)

    @abc.abstractmethod
    def key_partition(self,
                      topic: str,
                      key: Optional[bytes],
                      partition: int = None) -> Optional[int]:
        """Hash key to determine partition number."""
        ...

    def verify_recovery_event_path(self, now: float, tp: TP) -> None:
        ...

    @abc.abstractmethod
    def request_rejoin(self) -> None:
        """Force consumer to rejoin group."""
        ...

class ThreadDelegateConsumer(Consumer):

    _thread: ConsumerThread

    #: Main thread method queue.
    #: The consumer is running in a separate thread, and so we send
    #: requests to it via a queue.
    #: Sometimes the thread needs to call code owned by the main thread,
    #: such as App.on_partitions_revoked, and in that case the thread
    #: uses this method queue.
    _method_queue: MethodQueue

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._method_queue = MethodQueue(loop=self.loop, beacon=self.beacon)
        self.add_dependency(self._method_queue)
        self._thread = self._new_consumer_thread()
        self.add_dependency(self._thread)

    @abc.abstractmethod
    def _new_consumer_thread(self) -> ConsumerThread:
        ...

    async def threadsafe_partitions_revoked(
            self,
            receiver_loop: asyncio.AbstractEventLoop,
            revoked: Set[TP]) -> None:
        """Call rebalancing callback in a thread-safe manner."""
        promise = await self._method_queue.call(
            receiver_loop.create_future(),
            self.on_partitions_revoked,
            revoked,
        )
        # wait for main-thread to finish processing request
        await promise

    async def threadsafe_partitions_assigned(
            self,
            receiver_loop: asyncio.AbstractEventLoop,
            assigned: Set[TP]) -> None:
        """Call rebalancing callback in a thread-safe manner."""
        promise = await self._method_queue.call(
            receiver_loop.create_future(),
            self.on_partitions_assigned,
            assigned,
        )
        # wait for main-thread to finish processing request
        await promise

    async def _getmany(self,
                       active_partitions: Optional[Set[TP]],
                       timeout: float) -> RecordMap:
        return await self._thread.getmany(active_partitions, timeout)

    async def subscribe(self, topics: Iterable[str]) -> None:
        """Reset subscription (requires rebalance)."""
        await self._thread.subscribe(topics=topics)

    async def seek_to_committed(self) -> Mapping[TP, int]:
        """Seek all partitions to the committed offset."""
        return await self._thread.seek_to_committed()

    async def position(self, tp: TP) -> Optional[int]:
        """Return the current position for partition."""
        return await self._thread.position(tp)

    async def seek_wait(self, partitions: Mapping[TP, int]) -> None:
        """Seek partitions to specific offsets and wait."""
        return await self._thread.seek_wait(partitions)

    async def _seek(self, partition: TP, offset: int) -> None:
        self._thread.seek(partition, offset)

    def assignment(self) -> Set[TP]:
        """Return the current assignment."""
        return self._thread.assignment()

    def highwater(self, tp: TP) -> int:
        """Return the last available offset for specific partition."""
        return self._thread.highwater(tp)

    def topic_partitions(self, topic: str) -> Optional[int]:
        """Return the number of partitions configured for topic by name."""
        return self._thread.topic_partitions(topic)

    async def earliest_offsets(self, *partitions: TP) -> Mapping[TP, int]:
        """Return the earliest offsets for a list of partitions."""
        return await self._thread.earliest_offsets(*partitions)

    async def highwaters(self, *partitions: TP) -> Mapping[TP, int]:
        """Return the last offset for a list of partitions."""
        return await self._thread.highwaters(*partitions)

    async def _commit(self, offsets: Mapping[TP, int]) -> bool:
        return await self._thread.commit(offsets)

    def close(self) -> None:
        """Close consumer for graceful shutdown."""
        self._thread.close()

    def key_partition(self,
                      topic: str,
                      key: Optional[bytes],
                      partition: int = None) -> Optional[int]:
        """Hash key to determine partition number."""
        return self._thread.key_partition(topic, key, partition=partition)

    def verify_recovery_event_path(self, now: float, tp: TP) -> None:
        return self._thread.verify_recovery_event_path(now, tp)
    
    def request_rejoin(self):
        return self._thread.request_rejoin()



"""Terminal utilities."""
from mode.utils.compat import isatty
from .spinners import Spinner, SpinnerHandler
from .tables import Table, TableDataT, logtable, table

__all__ = [
    'Spinner',
    'SpinnerHandler',
    'Table',
    'TableDataT',
    'isatty',
    'logtable',
    'table',
]



import asyncio
from typing import Union
import pytest
from mode.utils.compat import want_bytes
from mode.utils.mocks import AsyncMock, Mock, call
from faust.livecheck import LiveCheck
from faust.livecheck.app import LiveCheckSensor
from faust.livecheck.exceptions import TestFailed
from faust.livecheck.locals import current_test_stack
from faust.livecheck.models import SignalEvent, TestExecution, TestReport
from faust.livecheck.signals import BaseSignal


class test_LiveCheckSensor:

    @pytest.fixture()
    def sensor(self):
        return LiveCheckSensor()

    def test_on_stream_event__no_test(self, *, sensor):
        stream = Mock()
        stream.current_test = None
        event = Mock()
        event.headers = {}
        state = sensor.on_stream_event_in(('topic', 'foo'), 3, stream, event)
        sensor.on_stream_event_out(('topic', 'foo'), 3, stream, event, state)

    def test_on_stream_event(self, *, sensor, execution):
        stream = Mock()
        stream.current_test = None
        event = Mock()
        event.headers = execution.as_headers()
        assert current_test_stack.top is None
        state = sensor.on_stream_event_in(('topic', 'foo'), 3, stream, event)
        assert current_test_stack.top.id == execution.id
        assert stream.current_test.id == execution.id
        sensor.on_stream_event_out(('topic', 'foo'), 3, stream, event, state)
        assert current_test_stack.top is None
        assert stream.current_test is None


class test_LiveCheck:

    @pytest.mark.parametrize('kwarg,value,expected_value', [
        ('test_topic_name', 'test-topic', 'test-topic'),
        ('bus_topic_name', 'bus-topic', 'bus-topic'),
        ('report_topic_name', 'report-topic', 'report-topic'),
        ('bus_concurrency', 1000, 1000),
        ('test_concurrency', 1001, 1001),
        ('send_reports', False, False),
    ])
    def test_constructor(self, kwarg, value, expected_value):
        app = LiveCheck('foo', **{kwarg: value})
        assert getattr(app, kwarg) == value

    def test_current_test(self, *, livecheck):
        test = Mock()
        assert livecheck.current_test is None
        with current_test_stack.push(test):
            assert livecheck.current_test is test
        assert livecheck.current_test is None

    def test__can_resolve(self, livecheck):
        assert isinstance(livecheck._can_resolve, asyncio.Event)

    def test_on_produce_attach_test_headers(
            self, *, livecheck, app, execution):
        headers = [('Foo-Bar-Baz', b'moo')]
        original_headers = list(headers)
        with current_test_stack.push(execution):
            livecheck.on_produce_attach_test_headers(
                sender=app,
                key=b'k',
                value=b'v',
                partition=3,
                headers=headers,
            )
            kafka_headers = {
                k: want_bytes(v) for k, v in execution.as_headers().items()
            }
            assert headers == (
                original_headers + list(kafka_headers.items()))

    def test_on_produce_attach_test_headers__no_test(self, *, livecheck, app):
        assert livecheck.current_test is None
        headers = []
        livecheck.on_produce_attach_test_headers(
            sender=app,
            key=b'k',
            value=b'v',
            partition=3,
            headers=headers,
        )
        assert not headers

    def test_on_produce_attach_test_headers__missing(self, *, livecheck, app):
        test = Mock()
        with current_test_stack.push(test):
            with pytest.raises(TypeError):
                livecheck.on_produce_attach_test_headers(
                    sender=app,
                    key=b'k',
                    value=b'v',
                    partition=3,
                    headers=None,
                )

    def test_case_decorator(self, *, livecheck):

        class SignalWithNoneOrigin(livecheck.Signal):
            __origin__ = None

        @livecheck.case()
        class test_foo:

            signal1: livecheck.Signal
            signal2: SignalWithNoneOrigin
            signal3: livecheck.Signal = livecheck.Signal()
            foo: Union[str, int]
            bar: str

        assert isinstance(test_foo.signal1, BaseSignal)
        assert isinstance(test_foo.signal2, SignalWithNoneOrigin)
        assert isinstance(test_foo.signal3, BaseSignal)

        assert test_foo.signal1.case is test_foo
        assert test_foo.signal2.case is test_foo
        assert test_foo.signal1.index == 1
        assert test_foo.signal2.index == 2
        assert test_foo.signal3.index == 3

    def test_add_case(self, *, livecheck):
        case = Mock()
        livecheck.add_case(case)
        assert livecheck.cases[case.name] is case

    @pytest.mark.asyncio
    async def test_post_report(self, *, livecheck):
        report = Mock()
        report.test = None
        livecheck.reports.send = AsyncMock()
        await livecheck.post_report(report)
        livecheck.reports.send.assert_called_once_with(key=None, value=report)

    @pytest.mark.asyncio
    async def test_post_report__with_test(self, *, livecheck):
        report = Mock()
        report.test = Mock()
        report.test.id = 'id'
        livecheck.reports.send = AsyncMock()
        await livecheck.post_report(report)
        livecheck.reports.send.assert_called_once_with(key='id', value=report)

    @pytest.mark.asyncio
    async def test_on_start(self, *, livecheck):
        livecheck._install_bus_agent = Mock()
        livecheck._install_test_execution_agent = Mock()

        await livecheck.on_start()

        livecheck._install_bus_agent.assert_called_once_with()
        livecheck._install_test_execution_agent.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_started(self, *, livecheck):
        case1 = Mock(name='case1')
        case2 = Mock(name='case2')
        livecheck.cases = {'a': case1, 'b': case2}
        livecheck.add_runtime_dependency = AsyncMock()

        await livecheck.on_started()

        livecheck.add_runtime_dependency.assert_has_calls([
            call.coro(case1),
            call.coro(case2),
        ])

    def test__install_bus_agent(self, *, livecheck):
        livecheck.agent = Mock()
        ag = livecheck._install_bus_agent()
        assert ag is livecheck.agent.return_value.return_value
        livecheck.agent.assert_called_once_with(
            channel=livecheck.bus,
            concurrency=livecheck.bus_concurrency,
        )

    def test__install_test_execution_agent(self, *, livecheck):
        livecheck.agent = Mock()
        ag = livecheck._install_test_execution_agent()
        assert ag is livecheck.agent.return_value.return_value
        livecheck.agent.assert_called_once_with(
            channel=livecheck.pending_tests,
            concurrency=livecheck.test_concurrency,
        )

    @pytest.mark.asyncio
    async def test__populate_signals(self, *, livecheck, execution):
        events = Mock()
        signal = SignalEvent(
            signal_name='foo',
            case_name=execution.case_name,
            key=b'k',
            value=b'v',
        )
        signal2 = SignalEvent(
            signal_name='foo',
            case_name='bar',
            key=b'k2',
            value=b'v2',
        )
        case = livecheck.cases[execution.case_name] = Mock(
            resolve_signal=AsyncMock(),
        )
        livecheck.cases.pop('bar', None)  # make sure 'bar' is missing

        async def iterate_events():
            yield execution.id, signal
            yield execution.id, signal2

        events.items.side_effect = iterate_events

        await livecheck._populate_signals(events)
        case.resolve_signal.assert_called_once_with(
            execution.id, signal)

    @pytest.mark.asyncio
    async def test__excecute_tests(self, *, livecheck, execution):
        tests = Mock()

        execution2 = execution.derive(case_name='bar')

        async def iterate_tests():
            yield execution.id, execution
            yield execution.id, execution2

        tests.items.side_effect = iterate_tests

        case = livecheck.cases[execution.case_name] = Mock(
            execute=AsyncMock(),
        )
        livecheck.cases.pop('bar', None)  # ensure 'bar' is missing.

        await livecheck._execute_tests(tests)
        case.execute.assert_called_once_with(execution)

    @pytest.mark.asyncio
    async def test__excecute_tests__raises(self, *, livecheck, execution):
        tests = Mock()

        async def iterate_tests():
            yield execution.id, execution

        tests.items.side_effect = iterate_tests

        case = livecheck.cases[execution.case_name] = Mock(
            execute=AsyncMock(side_effect=TestFailed()),
        )

        await livecheck._execute_tests(tests)
        case.execute.assert_called_once_with(execution)

    @pytest.mark.parametrize('name,origin,expected', [
        ('__main__.test_foo', 'examples.f.y', 'examples.f.y.test_foo'),
        ('examples.test_foo', 'examples.f.y', 'examples.test_foo'),
    ])
    def test_prepare_case_name(self, name, origin, expected, *, livecheck):
        livecheck.conf.origin = origin
        assert livecheck._prepare_case_name(name) == expected

    def test_prepare_case_name__no_origin(self, *, livecheck):
        livecheck.conf.origin = None
        with pytest.raises(RuntimeError):
            livecheck._prepare_case_name('__main__.test_foo')

    def test_bus(self, *, livecheck):
        livecheck.topic = Mock()
        assert livecheck.bus is livecheck.topic.return_value
        livecheck.topic.assert_called_once_with(
            livecheck.bus_topic_name,
            key_type=str,
            value_type=SignalEvent,
        )

    def test_pending_tests(self, *, livecheck):
        livecheck.topic = Mock()
        assert livecheck.pending_tests is livecheck.topic.return_value
        livecheck.topic.assert_called_once_with(
            livecheck.test_topic_name,
            key_type=str,
            value_type=TestExecution,
        )

    def test_reports(self, *, livecheck):
        livecheck.topic = Mock()
        assert livecheck.reports is livecheck.topic.return_value
        livecheck.topic.assert_called_once_with(
            livecheck.report_topic_name,
            key_type=str,
            value_type=TestReport,
        )



"""Tables (changelog stream)."""
import asyncio
import typing

from typing import Any, MutableMapping, Optional, Set, Tuple, cast

from mode import Service
from mode.utils.queues import ThrowableQueue

from faust.types import AppT, ChannelT, StoreT, TP
from faust.types.tables import CollectionT, TableManagerT
from faust.utils.tracing import traced_from_parent_span

from .recovery import Recovery

if typing.TYPE_CHECKING:
    from faust.app import App as _App
else:
    class _App: ...  # noqa

__all__ = [
    'TableManager',
]


class TableManager(Service, TableManagerT):
    """Manage tables used by Faust worker."""

    _channels: MutableMapping[CollectionT, ChannelT]
    _changelogs: MutableMapping[str, CollectionT]
    #: event that when set we cannot add any more tables.
    _tables_finalized: asyncio.Event
    #: event set when all tables have been registered
    _tables_registed: asyncio.Event
    #: event set when table recovery has started.
    _recovery_started: asyncio.Event
    _changelog_queue: Optional[ThrowableQueue]
    _pending_persisted_offsets: MutableMapping[TP, Tuple[StoreT, int]]

    _recovery: Optional[Recovery] = None

    def __init__(self, app: AppT, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.app = app
        self.data: MutableMapping = {}
        self._changelog_queue = None
        self._channels = {}
        self._changelogs = {}
        self._tables_finalized = asyncio.Event(loop=self.loop)
        self._tables_registered = asyncio.Event(loop=self.loop)
        self._recovery_started = asyncio.Event(loop=self.loop)

        self.actives_ready = False
        self.standbys_ready = False
        self._pending_persisted_offsets = {}

    def persist_offset_on_commit(self,
                                 store: StoreT,
                                 tp: TP,
                                 offset: int) -> None:
        """Mark the persisted offset for a TP to be saved on commit.

        This is used for "exactly_once" processing guarantee.
        Instead of writing the persisted offset to RocksDB when the message
        is sent, we write it to disk when the offset is committed.
        """
        existing_entry = self._pending_persisted_offsets.get(tp)
        if existing_entry is not None:
            _, existing_offset = existing_entry
            if offset < existing_offset:
                return
        self._pending_persisted_offsets[tp] = (store, offset)

    def on_commit(self, offsets: MutableMapping[TP, int]) -> None:
        """Call when committing source topic partitions."""
        # flush any pending persisted offsets added by
        # persist_offset_on_commit
        for tp in offsets:
            self.on_commit_tp(tp)

    def on_commit_tp(self, tp: TP) -> None:
        """Call when committing source topic partition used by this table."""
        entry = self._pending_persisted_offsets.get(tp)
        if entry is not None:
            store, offset = entry
            store.set_persisted_offset(tp, offset)

    def on_rebalance_start(self) -> None:
        """Call when a new rebalancing operation starts."""
        self.actives_ready = False
        self.standbys_ready = False

    def on_actives_ready(self) -> None:
        """Call when actives are fully up-to-date."""
        self.actives_ready = True

    def on_standbys_ready(self) -> None:
        """Call when standbys are fully up-to-date and ready for failover."""
        self.standbys_ready = True

    def __hash__(self) -> int:
        return object.__hash__(self)

    @property
    def changelog_topics(self) -> Set[str]:
        """Return the set of known changelog topics."""
        return set(self._changelogs.keys())

    @property
    def changelog_queue(self) -> ThrowableQueue:
        """Queue used to buffer changelog events."""
        if self._changelog_queue is None:
            self._changelog_queue = self.app.FlowControlQueue(
                maxsize=self.app.conf.stream_buffer_maxsize,
                loop=self.loop,
                clear_on_resume=True,
            )
        return self._changelog_queue

    @property
    def recovery(self) -> Recovery:
        """Recovery service used by this table manager."""
        if self._recovery is None:
            self._recovery = Recovery(
                self.app, self, beacon=self.beacon, loop=self.loop)
        return self._recovery

    def add(self, table: CollectionT) -> CollectionT:
        """Add table to be managed by this table manager."""
        if self._tables_finalized.is_set():
            raise RuntimeError('Too late to add tables at this point')
        assert table.name is not None
        if table.name in self:
            raise ValueError(f'Table with name {table.name!r} already exists')
        self[table.name] = table
        self._changelogs[table.changelog_topic.get_topic_name()] = table
        return table

    async def on_start(self) -> None:
        """Call when table manager is starting."""
        await self.sleep(1.0)
        if not self.should_stop:
            await self._update_channels()
            await self.recovery.start()

    async def wait_until_tables_registered(self) -> None:
        if not self.app.producer_only and not self.app.client_only:
            await self.wait_for_stopped(self._tables_registered)

    async def _update_channels(self) -> None:
        self._tables_finalized.set()
        for table in self.values():
            await asyncio.sleep(0)
            if table not in self._channels:
                chan = table.changelog_topic.clone_using_queue(
                    self.changelog_queue)
                self.app.topics.add(chan)
                await asyncio.sleep(0)
                self._channels[table] = chan
            await table.maybe_start()
        self.app.consumer.pause_partitions({
            tp for tp in self.app.consumer.assignment()
            if tp.topic in self._changelogs
        })
        await asyncio.sleep(0)
        self._tables_registered.set()

    async def on_stop(self) -> None:
        """Call when table manager is stopping."""
        await cast(_App, self.app)._fetcher.stop()
        if self._recovery:
            await self._recovery.stop()
        for table in self.values():
            await table.stop()

    def on_partitions_revoked(self, revoked: Set[TP]) -> None:
        """Call when cluster is rebalancing and partitions revoked."""
        T = traced_from_parent_span()
        T(self.recovery.on_partitions_revoked)(revoked)

    async def on_rebalance(self,
                           assigned: Set[TP],
                           revoked: Set[TP],
                           newly_assigned: Set[TP]) -> None:
        """Call when the cluster is rebalancing."""
        self._recovery_started.set()  # cannot add more tables.
        T = traced_from_parent_span()
        for table in self.values():
            await T(table.on_rebalance)(assigned, revoked, newly_assigned)

        await asyncio.sleep(0)
        await T(self._update_channels)()
        await asyncio.sleep(0)
        await T(self.recovery.on_rebalance)(assigned, revoked, newly_assigned)

    async def wait_until_recovery_completed(self) -> bool:
        if (self.recovery.started and not
                self.app.producer_only and not self.app.client_only):
            return await self.wait_for_stopped(self.recovery.completed)
        return False



import faust
import pytest


class Order(faust.Record, serializer='json'):
    account_id: str
    product_id: str
    amount: int
    price: float


app = faust.App('test-example')
orders_topic = app.topic('orders', value_type=Order)

# order count within the last hour (window is a 1-hour TumblingWindow).
orders_for_account = app.Table(
    'order-count-by-account', default=int,
).tumbling(3600).relative_to_stream()


@app.agent(orders_topic)
async def process_order(orders):
    async for order in orders.group_by(Order.account_id):
        orders_for_account[order.account_id] += 1
        yield order


@pytest.mark.asyncio()
async def test_process_order():
    app.finalize()
    app.conf.store = 'memory://'
    app.flow_control.resume()
    async with process_order.test_context() as agent:
        order = Order(account_id='1', product_id='2', amount=1, price=300)
        event = await agent.put(order)

        # windowed table: we select window relative to the current event
        assert orders_for_account['1'].current(event) == 1

        # in the window 3 hours ago there were no orders:
        assert not orders_for_account['1'].delta(3600 * 3, event)



"""Program ``faust model`` used to list details about a model."""
from datetime import datetime
from typing import Any, Sequence, Type

import click
from mode.utils import text

from faust.models import registry
from faust.types import FieldDescriptorT, ModelT
from faust.utils import terminal

from .base import AppCommand, argument

__all__ = ['model']

#: Built-in types show as name only (e.g. str).
#: Other types will use full `repr()`.
BUILTIN_TYPES = frozenset({int, float, str, bytes, datetime})


class model(AppCommand):
    """Show model detail."""

    headers = ['field', 'type', 'default']

    options = [
        argument('name'),
    ]

    async def run(self, name: str) -> None:
        """Dump list of registered models to terminal."""
        try:
            model = registry[name]
        except KeyError:
            if '.' in name:
                raise self._unknown_model(name)
            if self.app.conf.origin:
                lookup = '.'.join([self.app.conf.origin, name])
            else:
                lookup = name
            try:
                model = registry[lookup]
            except KeyError:
                raise self._unknown_model(name, lookup=lookup)
        self.say(
            self.tabulate(
                self.model_fields(model),
                headers=[self.bold(h) for h in self.headers],
                title=self._name(model),
                wrap_last_row=False,
            ))

    def _unknown_model(self, name: str, *,
                       lookup: str = None) -> click.UsageError:
        lookup = lookup or name
        alt = text.didyoumean(
            registry,
            lookup,
            fmt_none=f'Please run `{self.prog_name} models` for a list.')
        return click.UsageError(f'No model {name!r}. {alt}')

    def model_fields(self, model: Type[ModelT]) -> terminal.TableDataT:
        """Convert model fields to terminal table rows."""
        return [self.field(getattr(model, k)) for k in model._options.fields]

    def field(self, field: FieldDescriptorT) -> Sequence[str]:
        """Convert model field model to terminal table columns."""
        return [
            field.field,
            self._type(field.type),
            self.dark('*' if field.required else repr(field.default)),
        ]

    def _type(self, typ: Any) -> str:
        return typ.__name__ if typ in BUILTIN_TYPES else repr(typ)

    def model_to_row(self, model: Type[ModelT]) -> Sequence[str]:
        """Convert model to terminal table row."""
        return [
            self.bold_tail(self._name(model)),
            self.dark(self._help(model)),
        ]

    def _name(self, model: Type[ModelT]) -> str:
        return self.abbreviate_fqdn(model._options.namespace)

    def _help(self, model: Type[ModelT]) -> str:
        return model.__doc__ or '<N/A>'



import asyncio
import re
import faust
import pytest
from faust import Event, Record
from faust.exceptions import KeyDecodeError, ValueDecodeError
from faust.types import Message
from mode.utils.mocks import AsyncMock, Mock, call, patch


class Dummy(Record):
    foo: int


class test_Topic:

    @pytest.fixture
    def topic(self, *, app):
        return app.topic('foo')

    @pytest.fixture
    def topic_allow_empty(self, *, app):
        return app.topic('foo', allow_empty=True)

    @pytest.fixture
    def message(self):
        return Mock(name='message', autospec=Message)

    @pytest.fixture
    def message_empty_value(self):
        return Mock(name='message', value=None, headers=[], autospec=Message)

    def test_schema__default(self, *, topic):
        assert topic.key_type is None
        assert topic.value_type is None
        assert topic.key_serializer is None
        assert topic.value_serializer is None

        assert topic.schema is not None
        assert topic.schema.key_type is None
        assert topic.schema.value_type is None
        assert topic.schema.key_serializer is None
        assert topic.schema.value_serializer is None

        assert repr(topic.schema)

    def test_schema__from_schema(self, *, app):
        schema = faust.Schema(
            key_type='bytes',
            value_type='bytes',
            key_serializer='msgpack',
            value_serializer='msgpack',
        )
        topic = app.topic('foo', schema=schema)

        assert topic.key_type == 'bytes'
        assert topic.value_type == 'bytes'
        assert topic.key_serializer == 'msgpack'
        assert topic.value_serializer == 'msgpack'

        assert topic.schema is schema
        assert topic.schema.key_type == topic.key_type
        assert topic.schema.value_type == topic.value_type
        assert topic.schema.key_serializer == 'msgpack'
        assert topic.schema.value_serializer == 'msgpack'

        assert repr(topic.schema)

    def test_schema_loads_key__loads_arg_optional(self, *, app):
        topic = app.topic('foo', key_type='str', key_serializer='msgpack')
        app.serializers.loads_key = Mock(name='loads_value')

        message = Mock(name='message')
        payload = topic.schema.loads_key(app, message)

        app.serializers.loads_key.assert_called_once_with(
            'str', message.key, serializer='msgpack',
        )

        assert payload == app.serializers.loads_key.return_value

    def test_schema_loads_value__loads_arg_optional(self, *, app):
        topic = app.topic('foo', value_type='str', value_serializer='msgpack')
        app.serializers.loads_value = Mock(name='loads_value')

        message = Mock(name='message')
        payload = topic.schema.loads_value(app, message)

        app.serializers.loads_value.assert_called_once_with(
            'str', message.value, serializer='msgpack',
        )

        assert payload == app.serializers.loads_value.return_value

    def test_schema__overriding(self, *, app):
        schema = faust.Schema(
            key_type='bytes',
            value_type='bytes',
            key_serializer='msgpack',
            value_serializer='msgpack',
        )
        topic = app.topic(
            'foo',
            schema=schema,
            key_type='str',
            value_type='str',
            key_serializer='captnproto',
            value_serializer='captnproto',
        )

        assert topic.key_type == 'str'
        assert topic.value_type == 'str'
        assert topic.key_serializer == 'captnproto'
        assert topic.value_serializer == 'captnproto'
        assert topic.schema.key_type == 'str'
        assert topic.schema.value_type == 'str'
        assert topic.schema.key_serializer == 'captnproto'
        assert topic.schema.value_serializer == 'captnproto'

        assert repr(topic.schema)

    def test_init_key_serializer_taken_from_key_type(self, app):
        class M(Record, serializer='foobar'):
            x: int

        topic = app.topic('foo', key_type=M, value_type=M)
        assert topic.schema.key_serializer == 'foobar'
        assert topic.schema.value_serializer == 'foobar'

    @pytest.mark.asyncio
    async def test_publish_message__wait_enabled(self, *, topic, app):
        producer = Mock(send_and_wait=AsyncMock())
        topic._get_producer = AsyncMock(return_value=producer)
        callback = Mock(name='callback')
        headers = {'k': 'v'}
        fm = topic.as_future_message(
            key='foo',
            value='bar',
            partition=130,
            timestamp=312.5134,
            headers=headers,
            key_serializer='json',
            value_serializer='json',
            callback=callback,
        )
        await topic.publish_message(fm, wait=True)
        key, headers = topic.prepare_key('foo', 'json', None, headers)
        value, headers = topic.prepare_value('bar', 'json', None, headers)
        producer.send_and_wait.coro.assert_called_once_with(
            topic.get_topic_name(),
            key,
            value,
            headers=headers,
            partition=130,
            timestamp=312.5134,
        )
        callback.assert_called_once_with(fm)

    @pytest.mark.asyncio
    async def test_send__attachments_enabled(self, *, topic, app):
        app._attachments.enabled = True
        callback = Mock(name='callback')
        schema = Mock(name='schema')
        with patch('faust.topics.current_event') as current_event:
            await topic.send(
                key='k',
                value='v',
                partition=3,
                timestamp=312.41,
                headers={'k': 'v'},
                schema=schema,
                key_serializer='foo',
                value_serializer='bar',
                callback=callback,
                force=False,
            )
            current_event.assert_called_once_with()
            current_event()._attach.assert_called_once_with(
                topic,
                'k',
                'v',
                partition=3,
                timestamp=312.41,
                headers={'k': 'v'},
                schema=schema,
                key_serializer='foo',
                value_serializer='bar',
                callback=callback,
            )

    @pytest.mark.asyncio
    async def test_send__attachments_no_event(self, *, topic, app):
        app._attachments.enabled = True
        callback = Mock(name='callback')
        topic._send_now = AsyncMock()
        with patch('faust.topics.current_event', Mock(return_value=None)):
            await topic.send(
                key='k',
                value='v',
                partition=3,
                timestamp=312.41,
                headers={'k': 'v'},
                key_serializer='foo',
                value_serializer='bar',
                callback=callback,
                force=False,
            )
            topic._send_now.assert_called_once_with(
                'k',
                'v',
                partition=3,
                timestamp=312.41,
                headers={'k': 'v'},
                schema=None,
                key_serializer='foo',
                value_serializer='bar',
                callback=callback,
            )

    def test_on_published(self, *, topic, app):
        app.sensors.on_send_completed = Mock(name='on_send_completed')
        producer = Mock(name='producer')
        state = Mock(name='state')
        fut = Mock(name='fut', autospec=asyncio.Future)
        message = Mock(name='message', autospec=Message)
        topic._on_published(fut, message, producer, state)
        fut.result.assert_called_once_with()
        app.sensors.on_send_completed.assert_called_once_with(
            producer, state, fut.result())
        message.set_result.assert_called_once_with(fut.result())
        message.message.callback.assert_called_once_with(message)
        message.message.callback = None
        topic._on_published(fut, message, producer, state)

    def test_on_published__error(self, *, topic, app):
        app.sensors.on_send_error = Mock(name='on_send_error')
        producer = Mock(name='producer')
        state = Mock(name='state')
        fut = Mock(name='fut', autospec=asyncio.Future)
        exc = fut.result.side_effect = KeyError()
        message = Mock(name='message', autospec=Message)
        topic._on_published(fut, message, producer, state)

        message.set_exception.assert_called_once_with(exc)
        app.sensors.on_send_error.assert_called_once_with(
            producer, exc, state,
        )

    def test_aiter_when_iterator(self, *, topic):
        topic.is_iterator = True
        assert topic.__aiter__() is topic

    def test_send_soon(self, *, topic, app):
        topic.as_future_message = Mock(name='as_future_message')
        app.producer.send_soon = Mock(name='send_soon')
        callback = Mock(name='callback')
        schema = Mock(name='schema')
        fut = topic.send_soon(
            key=b'k',
            value=b'v',
            partition=3,
            timestamp=100.3,
            headers={'k': 'v'},
            schema=schema,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
        )
        topic.as_future_message.assert_called_once_with(
            key=b'k',
            value=b'v',
            partition=3,
            timestamp=100.3,
            headers={'k': 'v'},
            schema=schema,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
            eager_partitioning=False,
        )
        assert fut is topic.as_future_message.return_value

        app.producer.send_soon.assert_called_once_with(fut)

    @pytest.mark.asyncio
    async def test_decode__decode_error_propagate(self, *, topic, message):
        exc = KeyDecodeError()
        topic.app.serializers.loads_key = Mock(side_effect=exc)
        topic._compile_decode()
        with pytest.raises(type(exc)):
            await topic.decode(message, propagate=True)

    @pytest.mark.asyncio
    async def test_topic_schema_decode_helper(self, *, topic, message):
        await topic.schema.decode(topic.app, message)

    @pytest.mark.asyncio
    async def test_decode__decode_error_callback(self, *, topic, message):
        exc = KeyDecodeError()
        topic.app.serializers.loads_key = Mock(side_effect=exc)
        topic.on_key_decode_error = AsyncMock()
        topic._compile_decode()
        await topic.decode(message, propagate=False)
        topic.on_key_decode_error.assert_called_once_with(exc, message)

    @pytest.mark.asyncio
    async def test_derive(self, *, topic):
        t = topic.derive(key_serializer='raw')
        assert t.key_serializer == 'raw'
        assert t.value_serializer == topic.value_serializer
        assert t.topics == topic.topics

    @pytest.mark.asyncio
    async def test_derive__prefix(self, *, topic):
        topic.topics = ['foo', 'bar']
        t = topic.derive(key_serializer='raw', prefix='foo_')
        assert t.key_serializer == 'raw'
        assert t.value_serializer == topic.value_serializer
        assert t.topics == ['foo_foo', 'foo_bar']

    @pytest.mark.asyncio
    async def test_derive__with_pattern(self, *, topic):
        topic.topics = None
        topic.pattern = r'.*'
        with pytest.raises(ValueError):
            topic.derive(prefix='foo')

    @pytest.mark.asyncio
    async def test_decode_empty_error(self, *, topic, message_empty_value):
        topic = topic.derive_topic(value_serializer='json', value_type=Dummy)
        with pytest.raises(ValueDecodeError):
            await topic.decode(message_empty_value, propagate=True)

    @pytest.mark.asyncio
    async def test_decode_empty_error_callback(
            self, *, topic, message_empty_value):
        topic = topic.derive_topic(value_serializer='json', value_type=Dummy)
        topic.on_value_decode_error = AsyncMock()
        topic._compile_decode()
        await topic.decode(message_empty_value, propagate=False)
        topic.on_value_decode_error.assert_called_once()

    @pytest.mark.asyncio
    async def test_decode_allow_empty(self, *, topic_allow_empty,
                                      message_empty_value):
        event = await topic_allow_empty.decode(message_empty_value)
        assert event.value is None
        assert event.message == message_empty_value

    def test__topic_name_or_default__str(self, *, topic):
        assert topic._topic_name_or_default('xyz') == 'xyz'

    def test__topic_name_or_default__default(self, *, topic):
        assert topic._topic_name_or_default(None) == topic.get_topic_name()

    def test__topic_name_or_default__channel(self, *, topic, app):
        t2 = app.topic('bar')
        assert topic._topic_name_or_default(t2) == 'bar'

    @pytest.mark.asyncio
    async def test_maybe_declare(self, *, topic):
        topic.declare = AsyncMock()
        await asyncio.gather(
            topic.maybe_declare(),
            topic.maybe_declare(),
            topic.maybe_declare(),
            topic.maybe_declare(),
            topic.maybe_declare(),
            topic.maybe_declare(),
        )
        topic.declare.assert_called_once_with()
        await asyncio.gather(
            topic.maybe_declare(),
            topic.maybe_declare(),
            topic.maybe_declare(),
            topic.maybe_declare(),
            topic.maybe_declare(),
            topic.maybe_declare(),
        )
        assert topic.declare.call_count == 2

    @pytest.mark.asyncio
    async def test_declare__disabled(self, *, topic):
        topic.app.conf.topic_allow_declare = False
        producer = Mock(create_topic=AsyncMock())
        topic._get_producer = AsyncMock(return_value=producer)
        topic.partitions = 101
        topic.replicas = 202
        topic.topics = ['foo', 'bar']
        await topic.declare()
        producer.create_topic.coro.assert_not_called()

    @pytest.mark.asyncio
    async def test_declare(self, *, topic):
        topic.app.conf.topic_allow_declare = True
        producer = Mock(create_topic=AsyncMock())
        topic._get_producer = AsyncMock(return_value=producer)
        topic.partitions = 101
        topic.replicas = 202
        topic.topics = ['foo', 'bar']
        await topic.declare()
        producer.create_topic.coro.assert_has_calls([
            call(
                topic='foo',
                partitions=101,
                replication=202,
                config=topic.config,
                compacting=topic.compacting,
                deleting=topic.deleting,
                retention=topic.retention,
            ),
            call(
                topic='bar',
                partitions=101,
                replication=202,
                config=topic.config,
                compacting=topic.compacting,
                deleting=topic.deleting,
                retention=topic.retention,
            ),
        ])

    @pytest.mark.asyncio
    async def test_declare__defaults(self, *, topic):
        topic.app.conf.topic_allow_declare = True
        producer = Mock(create_topic=AsyncMock())
        topic._get_producer = AsyncMock(return_value=producer)
        topic.partitions = None
        topic.replicas = None
        topic.topics = ['foo', 'bar']
        await topic.declare()
        producer.create_topic.coro.assert_has_calls([
            call(
                topic='foo',
                partitions=topic.app.conf.topic_partitions,
                replication=topic.app.conf.topic_replication_factor,
                config=topic.config,
                compacting=topic.compacting,
                deleting=topic.deleting,
                retention=topic.retention,
            ),
            call(
                topic='bar',
                partitions=topic.app.conf.topic_partitions,
                replication=topic.app.conf.topic_replication_factor,
                config=topic.config,
                compacting=topic.compacting,
                deleting=topic.deleting,
                retention=topic.retention,
            ),
        ])

    @pytest.mark.asyncio
    async def test_put(self, *, topic):
        topic.is_iterator = True
        topic.queue.put = AsyncMock(name='queue.put')
        event = Mock(name='event', autospec=Event)
        await topic.put(event)
        topic.queue.put.assert_called_once_with(event)

    @pytest.mark.asyncio
    async def test_put__raise_when_not_iterator(self, *, topic):
        topic.is_iterator = False
        with pytest.raises(RuntimeError):
            await topic.put(Mock(name='event', autospec=Event))

    def test_set_pattern__raise_when_topics(self, *, topic):
        topic.topics = ['A', 'B']
        with pytest.raises(TypeError):
            topic.pattern = re.compile('something.*')

    def test_set_partitions__raise_when_zero(self, *, topic):
        with pytest.raises(ValueError):
            topic.partitions = 0

    def test_derive_topic__raise_when_no_sub(self, *m, topic):
        topic.topics = None
        topic.pattern = None
        with pytest.raises(TypeError):
            topic.get_topic_name()

    def test_derive_topic__raise_if_pattern_and_prefix(self, *, topic):
        topic.topics = None
        topic.pattern = re.compile('something2.*')
        with pytest.raises(ValueError):
            topic.derive_topic(suffix='-repartition')

    def test_get_topic_name__raise_when_pattern(self, *, topic):
        topic.topics = None
        topic.pattern = re.compile('^foo.$')
        with pytest.raises(TypeError):
            topic.get_topic_name()

    def test_get_topic_name__raise_if_multitopic(self, *, topic):
        topic.topics = ['t1', 't2']
        with pytest.raises(ValueError):
            topic.get_topic_name()



"""Faust environment variables."""
import os
from typing import Any, Sequence
from yarl import URL

__all__ = [
    'BLOCKING_TIMEOUT',
    'CONSOLE_PORT',
    'DATADIR',
    'DEBUG',
    'STRICT',
    'WEB_PORT',
    'WEB_BIND',
    'WEB_TRANSPORT',
    'WORKDIR',
    'SHUTDOWN_MEMORY_PERCENT',
    'SHUTDOWN_MEMORY_BYTES'
]

PREFICES: Sequence[str] = ['FAUST_', 'F_']


def _getenv(name: str, *default: Any,
            prefices: Sequence[str] = PREFICES) -> Any:
    for prefix in prefices:
        try:
            return os.environ[prefix + name]
        except KeyError:
            pass
    if default:
        return default[0]
    raise KeyError(prefices[0] + name)


#: Enables debugging features (like blockdetection).
DEBUG: bool = bool(_getenv('DEBUG', False))

#: Working directory to change into at start.
WORKDIR: str = _getenv('WORKDIR', None)

#: Directory to keep the application state (tables, checkpoints, etc).
DATADIR: str = _getenv('DATADIR', '{conf.name}-data')

#: Blocking detection timeout
BLOCKING_TIMEOUT: float = float(_getenv('BLOCKING_TIMEOUT', '10.0'))
FORCE_BLOCKING_TIMEOUT: bool = bool(_getenv('FORCE_BLOCKING_TIMEOUT', ''))

#: :pypi:`aiomonitor` console default port
CONSOLE_PORT: int = int(_getenv('CONSOLE_PORT', 50101))

STRICT: bool = bool(_getenv('STRICT', False))

WEB_PORT: int = int(_getenv('WEB_PORT', '6066'))
WEB_BIND: str = _getenv('F_WEB_BIND', '0.0.0.0')
WEB_TRANSPORT: URL = URL(_getenv('WEB_TRANSPORT', 'tcp://'))

SHUTDOWN_MEMORY_PERCENT: float = float(_getenv('SHUTDOWN_MEMORY_PERCENT', '0.0'))

SHUTDOWN_MEMORY_BYTES: float = float(_getenv('SHUTDOWN_MEMORY_BYTES', '0.0'))



"""LiveCheck Example.

1) First start an instance of the stock ordering system in a new terminal:

.. sourcecode:: console

    $ python examples/livecheck.py worker -l info

2) Then in a new terminal, start a LiveCheck instance for this app

.. sourcecode:: console

    $ python examples/livecheck.py livecheck -l info

3) Then visit ``http://localhost:6066/order/init/sell/`` in your browser.

    Alternatively you can use the ``post_order`` command:

    .. sourcecode:: console

        $ python examples/livecheck.py post_order --side=sell

The probability of a test execution happening is 50%
so have to do this at least twice to see activity happening
in the LiveCheck instance terminal.
"""
import asyncio
import faust
from faust import cli
from faust import web
from faust.livecheck import Case, Signal
from faust.types import FutureMessage, StreamT
from faust.utils import uuid


class Order(faust.Record):

    SIDE_SELL = 'sell'
    SIDE_BUY = 'buy'
    VALID_SIDES = {SIDE_SELL, SIDE_BUY}

    id: str
    user_id: str
    side: str
    quantity: float
    price: float

    # fake orders are not executed.
    fake: bool = False


app = faust.App(
    'orders',
    cache='redis://localhost:6379',
    origin='examples.livecheck',
    autodiscover=True,
)
livecheck = app.LiveCheck()

orders_topic = app.topic('orders', value_type=Order)
execution_topic = app.topic('order-execution', value_type=Order)


orders = web.Blueprint('orders')


@orders.route('/init/{side}/', name='init')
class OrderView(web.View):

    # First clients do a GET on /order/init/sell/
    # This endpoint will then do a POST to /order/create/

    async def get(self, request: web.Request, side: str) -> web.Response:
        order_id = uuid()
        user_id = uuid()
        side = side.lower()
        assert side in Order.VALID_SIDES

        # This will trigger our test_order case with 50% probability.
        # If executed we pass along LiveCheck-Test-* related headers
        # that can track the test as it progresses through the system.
        # All intermediate systems must pass along these headers,
        # be it through HTTP or Kafka.

        # we pass the side here as a testing "contract"
        # the test will ensure that no system is changing the side
        # of this order from buy to sell.
        async with test_order.maybe_trigger(order_id, side=side) as test:
            fake = bool(int(request.query.get('fake', 0)))
            next_url = self.url_for('orders:create', 'http://localhost:6066')
            data = {
                'order_id': order_id,
                'user_id': user_id,
                'side': side,
                'did_execute_test': bool(test),
                'fake': fake,
            }
            async with self.app.http_client.post(next_url, json=data) as resp:
                assert resp.status == 200
                return self.bytes(await resp.read(),
                                  content_type='application/json')


@orders.route('/create/', name='create')
class CreateOrderView(web.View):

    async def post(self, request: web.Request) -> web.Response:
        payload = await request.json()
        order_id = payload['order_id']
        user_id = payload['user_id']
        side = payload['side']
        fake = payload['fake']
        did_execute_test = payload['did_execute_test']

        if did_execute_test:
            # LiveCheck read the HTTP headers passed in this request
            # and set up a current_test environment.
            assert livecheck.current_test is not None
            # The id of the test execution should be the same as the order id.
            assert livecheck.current_test.id == order_id

        order = Order(order_id, user_id, side, 1.0, 3.33, fake=fake)
        await orders_topic.send(key=order_id, value=order)
        return self.json({'status': 'success'})


app.web.blueprints.add('/order/', orders)


@app.agent(orders_topic)
async def create_order(orders: StreamT[Order]) -> None:
    async for order in orders:
        test = livecheck.current_test
        if test is not None:
            assert test.id == order.id
        print('1. ORDER SENT TO DB')
        await test_order.order_sent_to_db.send(order)

        def on_order_sent(fut: FutureMessage) -> None:
            print('2. ORDER SENT TO KAFKA')
            asyncio.ensure_future(
                test_order.order_sent_to_kafka.send())

        await execution_topic.send(key=order.id, value=order,
                                   callback=on_order_sent)
        print('3. ORDER SENT TO EXECUTION AGENT')
        await app.cache.client.sadd(f'order.{order.user_id}.orders', order.id)
        await test_order.order_cache_in_redis.send()
        print('4. ORDER CACHED IN REDIS')


@app.agent(execution_topic)
async def execute_order(orders: StreamT[Order]) -> None:
    async for order in orders:
        execution_id = uuid()
        if not order.fake:
            # bla bla bla
            pass
        await test_order.order_executed.send(execution_id)
        print('5. ORDER EXECUTED BY EXECUTION AGENT')


@livecheck.case(warn_stalled_after=5.0, frequency=0.5, probability=0.5)
class test_order(Case):

    order_sent_to_db: Signal[Order]
    order_sent_to_kafka: Signal[None]
    order_cache_in_redis: Signal[None]
    order_executed: Signal[str]

    async def run(self, side: str) -> None:
        # 1) wait for order to be sent to database.
        order = await self.order_sent_to_db.wait(timeout=30.0)

        # contract:
        #   order id matches test execution id
        #   order.side matches test argument side.
        assert order.id == self.current_execution.test.id
        assert order.side == side

        # 2) wait for order to be sent to Kafka
        await self.order_sent_to_kafka.wait(timeout=30.0)

        # 3) wait for redis index to be updated.
        await self.order_cache_in_redis.wait(timeout=30.0)
        #  make sure it's now actually in redis
        assert await livecheck.cache.client.sismember(
            f'order.{order.user_id}.orders', order.id)

        # 4) wait for execution agent to execute the order.
        await self.order_executed.wait(timeout=30.0)

    async def make_fake_request(self) -> None:
        await self.get_url('http://localhost:6066/order/init/sell/?fake=1')


@app.command(
    cli.option('--side', default='sell', help='Order side: buy, sell'),
    cli.option('--base-url', default='http://localhost:6066'),
)
async def post_order(self: cli.AppCommand, side: str, base_url: str) -> None:
    path = self.app.web.url_for('orders:init', side=side)
    url = ''.join([base_url.rstrip('/'), path])
    async with self.app.http_client.get(url) as response:
        response.raise_for_status()
        print(await response.read())


if __name__ == '__main__':
    app.main()



import asyncio
import pytest
from faust.livecheck.exceptions import TestTimeout
from faust.livecheck.locals import current_execution_stack, current_test_stack
from faust.livecheck.models import SignalEvent
from faust.livecheck.signals import BaseSignal, Signal
from mode.utils.mocks import AsyncMock, Mock, patch


class test_BaseSignal:

    @pytest.fixture()
    def signal(self, *, case):
        return BaseSignal('foo', case, 1)

    @pytest.mark.asyncio
    async def test_send(self, *, signal):
        with pytest.raises(NotImplementedError):
            await signal.send('v', key='k')

    @pytest.mark.asyncio
    async def test_wait(self, *, signal):
        with pytest.raises(NotImplementedError):
            await signal.wait(key='k', timeout=10.0)

    @pytest.mark.asyncio
    async def test_resolve(self, *, signal):
        signal._set_current_value = Mock()
        signal._wakeup_resolvers = Mock()

        event = Mock()
        await signal.resolve('k', event)
        signal._set_current_value.assert_called_once_with('k', event)
        signal._wakeup_resolvers.assert_called_once_with()

    def test__set_name__(self, *, signal):
        signal.name = ''
        signal.__set_name__(type(signal), 'foo')
        assert signal.name == 'foo'
        signal.__set_name__(type(signal), 'bar')
        assert signal.name == 'foo'

    def test__wakeup_resolvers(self, *, signal):
        signal.case = Mock()
        signal._wakeup_resolvers()
        signal.case.app._can_resolve.set.assert_called_once_with()

    @pytest.mark.asyncio
    async def test__wait_for_resolved(self, *, signal):
        signal.case = Mock(app=Mock(wait=AsyncMock()))
        await signal._wait_for_resolved(timeout=3.01)
        signal.case.app.wait.assert_called_once_with(
            signal.case.app._can_resolve, timeout=3.01)
        signal.case.app._can_resolve.clear.assert_called_once_with()

    def test__get_current_value(self, *, signal):
        signal._set_current_value('k', 'foo')
        assert signal._get_current_value('k') == 'foo'

    def test__index_key(self, *, signal):
        assert signal._index_key('k') == (signal.name, signal.case.name, 'k')

    def test_repr(self, *, signal):
        assert repr(signal)


class test_Signal:

    @pytest.fixture()
    def signal(self, *, case):
        return Signal('foo', case, 1)

    @pytest.mark.asyncio
    async def test_send__no_test(self, *, signal):
        assert current_test_stack.top is None
        signal.case = Mock(app=Mock(bus=Mock(send=AsyncMock())))
        await signal.send('value', key='k', force=False)
        signal.case.app.bus.send.assert_not_called()

    @pytest.mark.asyncio
    async def test_send__no_test_force(self, *, signal):
        signal.case = Mock(app=Mock(bus=Mock(send=AsyncMock())))
        await signal.send('value', key='k', force=True)
        signal.case.app.bus.send.coro.assert_called_once_with(
            key='k',
            value=SignalEvent(
                signal_name=signal.name,
                case_name=signal.case.name,
                key='k',
                value='value',
            ),
        )

    @pytest.mark.asyncio
    async def test_send(self, *, signal, execution):
        with current_test_stack.push(execution):
            signal.case = Mock(app=Mock(bus=Mock(send=AsyncMock())))
            await signal.send('value', key=None, force=True)
            signal.case.app.bus.send.coro.assert_called_once_with(
                key=execution.id,
                value=SignalEvent(
                    signal_name=signal.name,
                    case_name=signal.case.name,
                    key=execution.id,
                    value='value',
                ),
            )

    @pytest.mark.asyncio
    async def test_wait__no_current_execution(self, *, signal):
        assert current_execution_stack.top is None
        with pytest.raises(RuntimeError):
            await signal.wait(timeout=1.01)

    @pytest.mark.asyncio
    async def test_wait(self, *, signal, runner):
        runner.on_signal_wait = AsyncMock()
        signal._wait_for_message_by_key = AsyncMock()
        runner.on_signal_received = AsyncMock()
        signal._verify_event = Mock()
        with current_execution_stack.push(runner):
            await signal.wait(timeout=2.02)
        signal._verify_event.assert_called_once()

    def test__verify_event(self, *, signal):
        ev = Mock()
        ev.key = 'key'
        ev.signal_name = 'name'
        ev.case_name = 'case'
        signal._verify_event(ev, 'key', 'name', 'case')

    @pytest.mark.asyncio
    async def test__wait_for_message_by_key__already_there(self, *, signal):
        signal._get_current_value = Mock(return_value='foo')
        ret = await signal._wait_for_message_by_key('k', timeout=1.0)
        assert ret == 'foo'

    @pytest.mark.asyncio
    async def test__wait_for_message_by_key__should_stop(self, *, signal):
        signal._get_current_value = Mock(side_effect=KeyError())
        app = signal.case.app
        app._stopped.set()
        with pytest.raises(asyncio.CancelledError):
            await signal._wait_for_message_by_key('k', timeout=1.0)

    @pytest.mark.asyncio
    async def test__wait_for_message_by_key(self, *, signal):
        signal._get_current_value = Mock(side_effect=KeyError())
        signal._wait_for_resolved = AsyncMock()

        async def on_resolved(**kwargs):
            if signal._wait_for_resolved.call_count >= 3:
                signal._get_current_value.side_effect = None
                signal._get_current_value.return_value = 30
        signal._wait_for_resolved.side_effect = on_resolved
        res = await signal._wait_for_message_by_key('k', timeout=1.0)
        assert res == 30

    @pytest.mark.asyncio
    async def test__wait_for_message_by_key__no_rem(self, *, signal):
        can_return = [False]
        signal._get_current_value = Mock()

        def on_get_value(k):
            if can_return[0]:
                signal._get_current_value.side_effect = None
                signal._get_current_value.return_value = 30
            raise KeyError()
        signal._get_current_value.side_effect = on_get_value

        with patch('faust.livecheck.signals.monotonic') as monotonic:
            monotonic.return_value = 111.1
            signal._wait_for_resolved = AsyncMock()

            async def on_resolved(**kwargs):
                if signal._wait_for_resolved.call_count >= 10:
                    monotonic.return_value = 333.3
                    can_return[0] = True

            signal._wait_for_resolved.side_effect = on_resolved
            res = await signal._wait_for_message_by_key('k', timeout=1.0)
            assert res == 30

    @pytest.mark.asyncio
    async def test__wait_for_message_by_key__no_timeout(self, *, signal):
        can_return = [False]
        signal._get_current_value = Mock()

        def on_get_value(k):
            if can_return[0]:
                signal._get_current_value.side_effect = None
                signal._get_current_value.return_value = 30
            raise KeyError()
        signal._get_current_value.side_effect = on_get_value

        with patch('faust.livecheck.signals.monotonic') as monotonic:
            monotonic.return_value = 111.1
            signal._wait_for_resolved = AsyncMock()

            async def on_resolved(**kwargs):
                if signal._wait_for_resolved.call_count >= 10:
                    monotonic.return_value = 333.3
                    can_return[0] = True

            signal._wait_for_resolved.side_effect = on_resolved
            res = await signal._wait_for_message_by_key('k', timeout=None)
            assert res == 30

    @pytest.mark.asyncio
    async def test__wait_for_message_by_key__timeout(self, *, signal):
        can_return = [False]
        signal._get_current_value = Mock(side_effect=KeyError())

        with patch('faust.livecheck.signals.monotonic') as monotonic:
            monotonic.return_value = 111.1
            signal._wait_for_resolved = AsyncMock()

            async def on_resolved(**kwargs):
                if signal._wait_for_resolved.call_count >= 10:
                    monotonic.return_value = 333.3
                    can_return[0] = True

            signal._wait_for_resolved.side_effect = on_resolved
            with pytest.raises(TestTimeout):
                await signal._wait_for_message_by_key('k', timeout=1.0)

    @pytest.mark.asyncio
    async def test__wait_for_message_by_key__app_stopped(self, *, signal):
        signal._get_current_value = Mock(side_effect=KeyError())

        with patch('faust.livecheck.signals.monotonic') as monotonic:
            monotonic.return_value = 111.1
            app = signal.case.app
            signal._wait_for_resolved = AsyncMock()

            async def on_resolved(**kwargs):
                app._stopped.set()

            signal._wait_for_resolved.side_effect = on_resolved
            with pytest.raises(asyncio.CancelledError):
                await signal._wait_for_message_by_key('k', timeout=1.0)



"""Copartitioned Assignor."""
from itertools import cycle
from math import ceil
from typing import Iterable, Iterator, MutableMapping, Optional, Sequence, Set
from mode.utils.typing import Counter
from .client_assignment import CopartitionedAssignment

__all__ = ['CopartitionedAssignor']


class CopartitionedAssignor:
    """Copartitioned Assignor.

    All copartitioned topics must have the same number of partitions

    The assignment is sticky which uses the following heuristics:

    - Maintain existing assignments as long as within capacity for each client
    - Assign actives to standbys when possible (within capacity)
    - Assign in order to fill capacity of the clients

    We optimize for not over utilizing resources instead of under-utilizing
    resources. This results in a balanced assignment when capacity is the
    default value which is ``ceil(num partitions / num clients)``

    Notes:
        Currently we raise an exception if number of clients is not enough
        for the desired `replication`.
    """

    capacity: int
    num_partitions: int
    replicas: int
    topics: Set[str]

    _num_clients: int
    _client_assignments: MutableMapping[str, CopartitionedAssignment]

    def __init__(self,
                 topics: Iterable[str],
                 cluster_asgn: MutableMapping[str, CopartitionedAssignment],
                 num_partitions: int,
                 replicas: int,
                 capacity: int = None) -> None:
        self._num_clients = len(cluster_asgn)
        assert self._num_clients, 'Should assign to at least 1 client'
        self.num_partitions = num_partitions
        self.replicas = min(replicas, self._num_clients - 1)
        self.capacity = (
            int(ceil(float(self.num_partitions) / self._num_clients))
            if capacity is None else capacity
        )
        self.topics = set(topics)

        assert self.capacity * self._num_clients >= self.num_partitions, \
            'Not enough capacity'

        self._client_assignments = cluster_asgn

    def get_assignment(self) -> MutableMapping[str, CopartitionedAssignment]:
        for copartitioned in self._client_assignments.values():
            copartitioned.unassign_extras(self.capacity, self.replicas)
        self._assign(active=True)
        self._assign(active=False)
        return self._client_assignments

    def _all_assigned(self, active: bool) -> bool:
        assigned_counts = self._assigned_partition_counts(active)
        total_assigns = self._total_assigns_per_partition(active)
        return all(assigned_counts[partition] == total_assigns
                   for partition in range(self.num_partitions))

    def _assign(self, active: bool) -> None:
        self._unassign_overassigned(active)
        unassigned = self._get_unassigned(active)
        self._assign_round_robin(unassigned, active)
        assert self._all_assigned(active)

    def _assigned_partition_counts(self, active: bool) -> Counter[int]:
        return Counter(
            partition
            for copartitioned in self._client_assignments.values()
            for partition in copartitioned.get_assigned_partitions(active)
        )

    def _get_client_limit(self, active: bool) -> int:
        return self.capacity * self._total_assigns_per_partition(active)

    def _total_assigns_per_partition(self, active: bool) -> int:
        return 1 if active else self.replicas

    def _unassign_overassigned(self, active: bool) -> None:
        # There are cases when multiple clients could have the same
        # assignment (zombies).  We need to handle that appropriately.
        partition_counts = self._assigned_partition_counts(active)
        total_assigns = self._total_assigns_per_partition(active=active)
        for partition in range(self.num_partitions):
            extras = partition_counts[partition] - total_assigns
            for _ in range(extras):
                assgn = next(
                    assgn
                    for assgn in self._client_assignments.values()
                    if assgn.partition_assigned(partition, active=active)
                )
                assgn.unassign_partition(partition, active=active)

    def _get_unassigned(self, active: bool) -> Sequence[int]:
        partition_counts = self._assigned_partition_counts(active)
        total_assigns = self._total_assigns_per_partition(active=active)
        assert all(
            partition_counts[partition] <= total_assigns
            for partition in range(self.num_partitions)
        )
        return [
            partition
            for partition in range(self.num_partitions)
            for _ in range(total_assigns - partition_counts[partition])
        ]

    def _can_assign(self, assignment: CopartitionedAssignment, partition: int,
                    active: bool) -> bool:
        return (
            not self._client_exhausted(assignment, active) and
            assignment.can_assign(partition, active)
        )

    def _client_exhausted(self, assignemnt: CopartitionedAssignment,
                          active: bool, client_limit: int = None) -> bool:
        if client_limit is None:
            client_limit = self._get_client_limit(active)
        return assignemnt.num_assigned(active) == client_limit

    def _find_promotable_standby(self, partition: int,
                                 candidates: Iterator[CopartitionedAssignment],
                                 ) -> Optional[CopartitionedAssignment]:
        # Round robin to find standby until we make a full cycle
        for _ in range(self._num_clients):
            assignment = next(candidates)
            can_assign = (
                assignment.partition_assigned(partition, active=False) and
                self._can_assign(assignment, partition, active=True)
            )
            if can_assign:
                return assignment
        return None

    def _find_round_robin_assignable(self, partition: int,
                                     candidates: Iterator[
                                         CopartitionedAssignment],
                                     active: bool,
                                     ) -> Optional[CopartitionedAssignment]:
        # Round robin and assign until we make a full circle
        for _ in range(self._num_clients):
            assignment = next(candidates)
            if self._can_assign(assignment, partition, active):
                return assignment
        return None

    def _assign_round_robin(self, unassigned: Iterable[int],
                            active: bool) -> None:
        # We do round robin assignment as follows:
        # - For actives, we first try to assign to a standby
        # - For standby, we offset the start for round robin to evenly
        # distribute standbys for colocated actives
        # - We do round robin
        # - If no assignment found, it must be a standby and the only unfilled
        # client(s) must be actives/standbys for the partition
        # - If no assignment found, we unassign and arbitrary partition from a
        # filled assignment such that the partition can be assigned to it
        # - This guarantees eventual assignment of all partitions
        client_limit = self._get_client_limit(active)
        candidates = cycle(self._client_assignments.values())
        unassigned = list(unassigned)
        while unassigned:
            partition = unassigned.pop(0)

            assign_to = None

            if active:
                # For actives we first try to find a standby to assign to
                assign_to = self._find_promotable_standby(partition,
                                                          candidates)
                if assign_to is not None:
                    # Unassign standby which will be promoted
                    assign_to.unassign_partition(partition, active=False)
            else:
                # For standbys we offset to round robin start to shuffle
                # assignment of standbys
                for _ in range(partition):
                    next(candidates)

            assert assign_to is None or active

            assign_to = assign_to or self._find_round_robin_assignable(
                partition, candidates, active)

            # If round robin assignment didn't work then we must be
            # assigning a standby and the only un-exhausted clients
            # are actives for the partition
            assert (
                assign_to is not None or
                (
                    not active and
                    all(
                        assgn.partition_assigned(partition, active=True) or
                        assgn.partition_assigned(partition, active=False) or
                        self._client_exhausted(assgn, active, client_limit)
                        for assgn in self._client_assignments.values()
                    )
                )
            )

            # If round robin didn't work, we free up the first full standby
            # assignment to which the partition can be assigned.
            if assign_to is None:
                assign_to = next(
                    assigment
                    for assigment in self._client_assignments.values()
                    if (self._client_exhausted(assigment, active) and
                        assigment.can_assign(partition, active))
                )  # By above assertion, should never throw error
                unassigned_partition = assign_to.pop_partition(active)
                unassigned.append(unassigned_partition)

            # Assign partition
            assign_to.assign_partition(partition, active)



"""Table (key/value changelog stream)."""
from typing import Any, ClassVar, Type

from mode import Seconds

from faust import windows
from faust.types.tables import KT, TableT, VT, WindowWrapperT
from faust.types.windows import WindowT
from faust.types.tuples import MessageSentCallback
from faust.utils.terminal.tables import dict_as_ansitable

from . import wrappers
from .base import Collection

__all__ = ['Table']


class Table(TableT[KT, VT], Collection):
    """Table (non-windowed)."""

    WindowWrapper: ClassVar[Type[WindowWrapperT]] = wrappers.WindowWrapper

    def using_window(self, window: WindowT, *,
                     key_index: bool = False) -> WindowWrapperT:
        """Wrap table using a specific window type."""
        self.window = window
        self._changelog_compacting = True
        self._changelog_deleting = True
        self._changelog_topic = None  # will reset on next property access
        return self.WindowWrapper(self, key_index=key_index)

    def hopping(self, size: Seconds, step: Seconds,
                expires: Seconds = None,
                key_index: bool = False) -> WindowWrapperT:
        """Wrap table in a hopping window."""
        return self.using_window(
            windows.HoppingWindow(size, step, expires),
            key_index=key_index,
        )

    def tumbling(self, size: Seconds,
                 expires: Seconds = None,
                 key_index: bool = False) -> WindowWrapperT:
        """Wrap table in a tumbling window."""
        return self.using_window(
            windows.TumblingWindow(size, expires),
            key_index=key_index,
        )

    def __missing__(self, key: KT) -> VT:
        if self.default is not None:
            return self.default()
        raise KeyError(key)

    def _has_key(self, key: KT) -> bool:
        return key in self

    def _get_key(self, key: KT) -> VT:
        return self[key]

    def _set_key(self, key: KT, value: VT) -> None:
        self[key] = value

    def _del_key(self, key: KT) -> None:
        del self[key]

    def on_key_get(self, key: KT, partition: int = None) -> None:
        """Call when the value for a key in this table is retrieved."""
        self._sensor_on_get(self, key)

    def on_key_set(self, 
                   key: KT, 
                   value: VT, 
                   partition: int = None, 
                   callback: MessageSentCallback = None) -> None:
        """Call when the value for a key in this table is set."""
        if partition is None:
            partition = self.partition_for_key(key)
        fut = self.send_changelog(partition, key, value, callback=callback)
        # partition may be None, in which case the finalized partition
        # is in fut.partition
        partition = fut.message.partition
        assert partition is not None
        self._maybe_set_key_ttl(key, partition)
        self._sensor_on_set(self, key, value)

    def on_key_del(self, 
                   key: KT, 
                   partition: int = None, 
                   callback: MessageSentCallback = None) -> None:
        """Call when a key in this table is removed."""
        if partition is None:
            partition = self.partition_for_key(key)        
        fut = self.send_changelog(partition, key, value=None,
                                  value_serializer='raw', 
                                  callback=callback)
        partition = fut.message.partition
        assert partition is not None
        self._maybe_del_key_ttl(key, partition)
        self._sensor_on_del(self, key)

    def as_ansitable(self, title: str = '{table.name}',
                     **kwargs: Any) -> str:
        """Draw table as a a terminal ANSI table."""
        return dict_as_ansitable(
            self,
            title=title.format(table=self),
            **kwargs)



import asyncio
import pytest

from mode import label, shortlabel
from mode.utils.futures import done_future
from mode.utils.mocks import AsyncMock, Mock, patch

from faust import App, Channel, Topic
from faust.transport.consumer import Consumer
from faust.transport.conductor import Conductor
from faust.types import Message, TP

TP1 = TP('foo', 0)
TP2 = TP('foo', 1)


class test_Conductor:

    @pytest.fixture
    def con(self, *, app):
        return Conductor(app)

    @pytest.fixture
    def con_client_only(self, *, app):
        app.client_only = True
        return Conductor(app)

    def test_constructor(self, *, con):
        assert con._topics == set()
        assert con._topic_name_index == {}
        assert con._tp_index == {}
        assert con._tp_to_callback == {}
        assert con._acking_topics == set()
        assert con._subscription_changed is None
        assert con._subscription_done is None
        assert con._compiler
        assert con.on_message

    def test_acks_enabled_for(self, *, con):
        assert not con.acks_enabled_for('foo')
        con._acking_topics.add('foo')
        assert con.acks_enabled_for('foo')

    @pytest.mark.asyncio
    async def test_con_client_only(self, *, con_client_only):
        assert con_client_only.on_message
        message = Mock(name='message')
        tp = TP(topic=message.topic, partition=0)
        cb = con_client_only._tp_to_callback[tp] = AsyncMock(name='cb')

        ret = await con_client_only.on_message(message)
        assert ret is cb.coro.return_value

        cb.assert_called_once_with(message)

    @pytest.mark.asyncio
    async def test_commit(self, *, con):
        con.app = Mock(
            name='app',
            autospec=App,
            consumer=Mock(
                autospec=Consumer,
                commit=AsyncMock(),
            ),
        )
        await con.commit({TP1})

        con.app.consumer.commit.assert_called_once_with({TP1})

    @pytest.mark.asyncio
    async def test_on_message(self, *, con):
        cb = con._tp_to_callback[TP1] = AsyncMock(name='callback')
        message = Mock(name='message', autospec=Message)
        message.tp = TP1
        await con.on_message(message)
        cb.assert_called_once_with(message)

    @pytest.mark.asyncio
    async def test_wait_for_subscriptions(self, *, con):
        with patch('asyncio.Future', AsyncMock()) as Future:
            con._subscription_done = None
            await con.wait_for_subscriptions()
            Future.assert_called_once_with(loop=con.loop)

    @pytest.mark.asyncio
    async def test_wait_for_subscriptions__done(self, *, con):
        con._subscription_done = done_future()
        await con.wait_for_subscriptions()

    @pytest.mark.asyncio
    async def test_update_indices(self, *, con):
        topic1 = Mock(name='topic1', autospec=Topic)
        topic1.acks = False
        topic1.topics = ['t1']
        topic1.internal = False
        topic2 = Mock(name='topic2', autospec=Topic)
        topic2.acks = True
        topic2.topics = ['t2']
        topic2.internal = True
        topic2.maybe_declare = AsyncMock(name='maybe_declare')
        con._topics = {topic1, topic2}

        await con._update_indices()
        topic1.maybe_declare.assert_not_called()
        topic2.maybe_declare.assert_called_once_with()
        assert 't1' not in con._acking_topics
        assert 't2' in con._acking_topics
        assert con._topic_name_index['t1'] == {topic1}
        assert con._topic_name_index['t2'] == {topic2}

    @pytest.mark.asyncio
    async def test_on_partitions_assigned(self, *, con):
        con._tp_index = {1: 2}
        con._update_tp_index = Mock(name='_update_tp_index')
        con._update_callback_map = Mock(name='_update_callback_map')
        assigned = {TP1, TP2}
        await con.on_partitions_assigned(assigned)
        assert not con._tp_index
        con._update_tp_index.assert_called_once_with(assigned)
        con._update_callback_map.assert_called_once_with()

    def test_update_tp_index(self, *, con):
        assigned = {TP1, TP2}
        topic1 = Mock(name='topic1', autospec=Topic)
        topic1.topics = [TP1.topic]
        topic1.active_partitions = None
        con._topics.add(topic1)
        con._update_tp_index(assigned)
        assert topic1 in con._tp_index[TP1]
        assert topic1 in con._tp_index[TP2]

    def test_update_tp_index__active_partitions(self, *, con):
        assigned = {TP1, TP2}
        topic1 = Mock(name='topic1', autospec=Topic)
        topic1.active_partitions = {TP1, TP2}
        con._topics.add(topic1)
        con._update_tp_index(assigned)
        assert topic1 in con._tp_index[TP1]
        assert topic1 in con._tp_index[TP2]
        con._update_tp_index(set())

    def test_update_tp_index__active_partitions_empty(self, *, con):
        assigned = {TP1, TP2}
        topic1 = Mock(name='topic1', autospec=Topic)
        topic1.active_partitions = set()
        con._topics.add(topic1)
        con._update_tp_index(assigned)
        assert topic1 not in con._tp_index[TP1]
        assert topic1 not in con._tp_index[TP2]

    def test_update_callback_map(self, *, con):
        chan1 = Mock(name='chan1', autospec=Channel)
        chan2 = Mock(name='chan2', autospec=Channel)
        con._tp_index = {
            TP1: chan1,
            TP2: chan2,
        }
        con._update_callback_map()

        assert con._tp_to_callback[TP1]
        assert con._tp_to_callback[TP2]

    def test_clear(self, *, con):
        con._topics = {'t1'}
        con._topic_name_index = {2: 3}
        con._tp_index = {3: 4}
        con._tp_to_callback = {4: 5}
        con._acking_topics = {1, 2, 3}
        con.clear()

        assert not con._topics
        assert not con._topic_name_index
        assert not con._tp_index
        assert not con._tp_to_callback
        assert not con._acking_topics

    def test_iter(self, *, con):
        con._topics = {'1', '2'}
        assert sorted(iter(con)) == ['1', '2']

    def test_hash(self, *, con):
        assert hash(con)

    def test_add(self, *, app, con):
        con._topics = set()
        con._topic_name_index = {}
        topic = app.topic('foo')
        con.add(topic)
        assert topic in con._topics
        con.add(topic)
        con.add(app.topic('bar'))

    def test_topic_contain_unsubcribed_topics(self, *, app, con):
        con._topic_name_index = {'foo': {}}
        assert con._topic_contain_unsubscribed_topics(app.topic('bar'))
        assert not con._topic_contain_unsubscribed_topics(app.topic('foo'))

    def test_flag_changes(self, *, con):
        con._subscription_changed = None
        con._subscription_done = None
        con._flag_changes()
        assert con._subscription_done is not None

        con._subscription_done = None
        con._subscription_changed = asyncio.Event()
        con._flag_changes()
        assert con._subscription_changed.is_set()
        assert con._subscription_done is not None
        con._flag_changes()

    def test_add_flags_changes(self, *, con, app):
        topic = app.topic('foo', 'bar')
        con._flag_changes = Mock(name='flag_changes')
        con._topic_name_index['baz'].add(topic)
        con.add(topic)
        con._flag_changes.assert_called_once_with()

    def test_discard(self, *, con, app):
        topic = app.topic('foo', 'bar')
        con.add(topic)
        assert topic in con._topics
        con.discard(topic)
        assert topic not in con._topics

    def test_label(self, *, con):
        assert label(con)

    @pytest.mark.asyncio
    async def test_wait_for_subscriptions__notset(self, *, con):
        with patch('asyncio.Future') as Future:
            Future.return_value = done_future()
            await con.wait_for_subscriptions()

    @pytest.mark.asyncio
    async def test_maybe_wait_for_subscriptions(self, *, con):
        con._subscription_done = done_future()
        await con.maybe_wait_for_subscriptions()

    @pytest.mark.asyncio
    async def test_maybe_wait_for_subscriptions__none(self, *, con):
        con._subscription_done = None
        await con.maybe_wait_for_subscriptions()

    @pytest.mark.asyncio
    async def test_on_client_only_start(self, *, con, app):
        topic = app.topic('foo', 'bar')
        con.add(topic)
        await con.on_client_only_start()
        assert con._tp_index[TP(topic='foo', partition=0)] == {topic}
        assert con._tp_index[TP(topic='bar', partition=0)] == {topic}

    def test_shortlabel(self, *, con):
        assert shortlabel(con)



import faust

__all__ = ['Error', 'Status']


class Error(faust.Record):

    #: Message (the actual formatted log message).
    message: str

    #: Format (sent to logging, e.g. 'the %s for %s did %r'
    format: str   # noqa

    hostname: str = None

    #: Traceback (if any)
    traceback: str = None

    #: Name of the origin logger.
    logger: str = None

    #: Path to the file logging this.
    filename: str = None

    #: Name of module logging this.
    module: str = None

    #: Line number
    lineno: int = None

    #: Logging severity
    severity: str = 'ERROR'

    timestamp: float = None

    app_id: str = None


class Status(faust.Record):

    #: The id of the app that is sending this.
    app_id: str

    #: Worker hostname
    hostname: str

    #: What test is being reported on.
    category: str

    #: What state is this test in
    state: str

    #: What color should we display this as
    color: str

    #: How many times did this fail so far.
    count: int

    #: Severity of issue
    severity: str

    @property
    def key(self):
        return (self.app_id, self.hostname, self.category)

    @property
    def details(self):
        return {
            'state': self.state,
            'count': self.count,
            'severity': self.severity,
            'color': self.color,
        }



import pytest
from faust.tables.recovery import Recovery
from faust.types import TP
from mode.utils.mocks import AsyncMock, Mock

TP1 = TP('foo', 3)
TP2 = TP('bar', 4)
TP3 = TP('baz', 5)


class test_Manager:

    @pytest.fixture()
    def tables(self, *, app):
        return app.tables

    def test_persist_offset_on_commit(self, *, tables):
        store = Mock(name='store')
        tables.persist_offset_on_commit(store, TP1, 30)
        assert tables._pending_persisted_offsets[TP1] == (store, 30)

        tables.persist_offset_on_commit(store, TP1, 29)
        assert tables._pending_persisted_offsets[TP1] == (store, 30)

        tables.persist_offset_on_commit(store, TP1, 31)
        assert tables._pending_persisted_offsets[TP1] == (store, 31)

    def test_on_commit(self, *, tables):
        tables.on_commit_tp = Mock(name='on_commit_tp')
        tables.on_commit({TP1: 30})
        tables.on_commit_tp.assert_called_once_with(TP1)

    def test_on_commit_tp(self, *, tables):
        store = Mock(name='store')
        tables.on_commit_tp(TP1)
        tables.persist_offset_on_commit(store, TP1, 30)
        tables.on_commit_tp(TP1)
        store.set_persisted_offset.assert_called_once_with(TP1, 30)

    def test_on_rebalance_start(self, *, tables):
        tables.on_rebalance_start()
        assert not tables.actives_ready
        assert not tables.standbys_ready

        tables.on_actives_ready()
        assert tables.actives_ready

        tables.on_standbys_ready()
        assert tables.standbys_ready

    def test_hash(self, *, tables):
        assert hash(tables)

    def test_changelog_topics(self, *, tables):
        assert tables.changelog_topics == set()

    def test_changelog_queue(self, *, tables, app):
        assert tables.changelog_queue
        assert tables.changelog_queue.maxsize == app.conf.stream_buffer_maxsize

    def test_recovery(self, *, tables):
        assert tables.recovery
        assert isinstance(tables.recovery, Recovery)
        assert tables.recovery.beacon.parent is tables.beacon
        assert tables.recovery.loop is tables.loop

    def test_add(self, *, tables):
        table = Mock(name='table')
        assert tables.add(table) is table
        assert tables[table.name] is table
        assert tables._changelogs[table.changelog_topic.get_topic_name()]

        with pytest.raises(ValueError):
            tables.add(table)  # already exists

    def test_add__when_recovery_started_raises(self, *, tables):
        tables._tables_finalized.set()
        with pytest.raises(RuntimeError):
            tables.add(Mock(name='table'))

    @pytest.mark.asyncio
    async def test_on_start(self, *, tables):
        tables.sleep = AsyncMock()
        tables._update_channels = AsyncMock()
        tables._recovery = Mock(start=AsyncMock())

        tables._stopped.set()
        await tables.on_start()

        tables.sleep.assert_called_once_with(1.0)
        tables._update_channels.assert_not_called()
        tables._recovery.start.assert_not_called()

        tables._stopped.clear()
        await tables.on_start()

        tables._update_channels.assert_called_once_with()
        tables._recovery.start.assert_called_once_with()

    @pytest.mark.asyncio
    async def test__update_channels(self, *, tables, app):
        app.consumer = Mock(name='consumer')
        app.consumer.assignment.return_value = set()
        await tables._update_channels()
        tables._tables_finalized.clear()

        app.topics = Mock(name='topics')
        table1 = Mock(name='table', maybe_start=AsyncMock())
        tables.add(table1)
        await tables._update_channels()
        tables._tables_finalized.clear()

        table1.maybe_start.assert_called_once_with()
        assert tables._channels[table1]
        await tables._update_channels()
        assert tables._tables_finalized.is_set()

    @pytest.mark.asyncio
    async def test_on_stop(self, *, tables, app):
        app._fetcher = Mock(name='fetcher', stop=AsyncMock())
        await tables.on_stop()

        app._fetcher.stop.assert_called_once_with()

        table1 = Mock(name='table1', stop=AsyncMock())
        tables._recovery = Mock(name='_recovery', stop=AsyncMock())
        tables.add(table1)
        await tables.on_stop()

        tables._recovery.stop.assert_called_once_with()
        table1.stop.assert_called_once_with()

    def test_on_partitions_revoked(self, *, tables):
        tables._recovery = Mock(name='_recovery')
        tables.on_partitions_revoked({TP1})
        tables._recovery.on_partitions_revoked.assert_called_once_with({TP1})

    @pytest.mark.asyncio
    async def test_on_rebalance(self, *, tables):
        tables._update_channels = AsyncMock()
        tables._recovery = Mock(on_rebalance=AsyncMock())

        await tables.on_rebalance({TP1, TP2, TP3}, set(), {TP1, TP2, TP3})
        tables._recovery_started.clear()

        tables._update_channels.assert_called_once_with()
        tables._recovery.on_rebalance.assert_called_once_with(
            {TP1, TP2, TP3}, set(), {TP1, TP2, TP3})

        table1 = Mock(name='table', on_rebalance=AsyncMock())
        tables.add(table1)

        await tables.on_rebalance({TP2, TP3}, {TP1}, set())
        table1.on_rebalance.assert_called_once_with({TP2, TP3}, {TP1}, set())
